"""Unit tests for Lie algebra."""

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.skew_symmetric_matrices import SkewSymmetricMatrices


class TestLieAlgebra(geomstats.tests.TestCase):
    def setUp(self):
        self.n = 4
        self.dim = int(self.n * (self.n - 1) / 2)
        self.algebra = SkewSymmetricMatrices(n=self.n)

    def test_dimension(self):
        result = self.algebra.dim
        expected = self.dim
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_matrix_representation_and_belongs(self):
        n_samples = 2
        point = gs.random.rand(n_samples * self.dim)
        point = gs.reshape(point, (n_samples, self.dim))
        mat = self.algebra.matrix_representation(point)
        result = self.algebra.belongs(mat).all()
        expected = True
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_basis_and_matrix_representation(self):
        n_samples = 2
        expected = gs.random.rand(n_samples * self.dim)
        expected = gs.reshape(expected, (n_samples, self.dim))
        mat = self.algebra.matrix_representation(expected)
        result = self.algebra.basis_representation(mat)
        self.assertAllClose(result, expected)



"""Numpy based linear algebra backend."""

import autograd.numpy as np
import scipy.linalg
from autograd.numpy.linalg import (  # NOQA
    det,
    eig,
    eigh,
    eigvalsh,
    inv,
    norm,
    svd
)

from .common import to_ndarray

_TOL = 1e-10


def _is_symmetric(x, tol=_TOL):
    new_x = to_ndarray(x, to_ndim=3)
    return (np.abs(new_x - np.transpose(new_x, axes=(0, 2, 1))) < tol).all()


def _expsym(x):
    eigvals, eigvecs = np.linalg.eigh(x)
    eigvals = np.exp(eigvals)
    eigvals = np.vectorize(np.diag, signature='(n)->(n,n)')(eigvals)
    transp_eigvecs = np.transpose(eigvecs, axes=(0, 2, 1))
    result = np.matmul(eigvecs, eigvals)
    result = np.matmul(result, transp_eigvecs)
    return result


def expm(x):
    ndim = x.ndim
    new_x = to_ndarray(x, to_ndim=3)
    if _is_symmetric(new_x):
        result = _expsym(new_x)
    else:
        result = np.vectorize(scipy.linalg.expm,
                              signature='(n,m)->(n,m)')(new_x)

    if ndim == 2:
        return result[0]
    return result


def logm(x):
    ndim = x.ndim
    new_x = to_ndarray(x, to_ndim=3)
    if _is_symmetric(new_x):
        eigvals, eigvecs = np.linalg.eigh(new_x)
        if (eigvals > 0).all():
            eigvals = np.log(eigvals)
            eigvals = np.vectorize(np.diag, signature='(n)->(n,n)')(eigvals)
            transp_eigvecs = np.transpose(eigvecs, axes=(0, 2, 1))
            result = np.matmul(eigvecs, eigvals)
            result = np.matmul(result, transp_eigvecs)
        else:
            result = np.vectorize(scipy.linalg.logm,
                                  signature='(n,m)->(n,m)')(new_x)
    else:
        result = np.vectorize(scipy.linalg.logm,
                              signature='(n,m)->(n,m)')(new_x)

    if ndim == 2:
        return result[0]
    return result


def powerm(x, power):
    ndim = x.ndim
    new_x = to_ndarray(x, to_ndim=3)
    if _is_symmetric(new_x):
        eigvals, eigvecs = np.linalg.eigh(new_x)
        if (eigvals > 0).all():
            eigvals = eigvals ** power
            eigvals = np.vectorize(np.diag, signature='(n)->(n,n)')(eigvals)
            transp_eigvecs = np.transpose(eigvecs, axes=(0, 2, 1))
            result = np.matmul(eigvecs, eigvals)
            result = np.matmul(result, transp_eigvecs)
        else:
            log_x = np.vectorize(scipy.linalg.logm,
                                 signature='(n,m)->(n,m)')(new_x)
            p_log_x = power * log_x
            result = np.vectorize(scipy.linalg.expm,
                                  signature='(n,m)->(n,m)')(p_log_x)
    else:
        log_x = np.vectorize(scipy.linalg.logm,
                             signature='(n,m)->(n,m)')(new_x)
        p_log_x = power * log_x
        result = np.vectorize(scipy.linalg.expm,
                              signature='(n,m)->(n,m)')(p_log_x)

    if ndim == 2:
        return result[0]
    return result


def sqrtm(x):
    return np.vectorize(
        scipy.linalg.sqrtm, signature='(n,m)->(n,m)')(x)


def qr(*args, **kwargs):
    return np.vectorize(np.linalg.qr,
                        signature='(n,m)->(n,k),(k,m)',
                        excluded=['mode'])(*args, **kwargs)



"""Unit tests for Riemannian KMeans."""

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry import hypersphere
from geomstats.geometry import spd_matrices
from geomstats.learning.frechet_mean import FrechetMean
from geomstats.learning.kmeans import RiemannianKMeans


class TestRiemannianKMeans(geomstats.tests.TestCase):
    _multiprocess_can_split_ = True

    @geomstats.tests.np_and_pytorch_only
    def test_hypersphere_kmeans_fit(self):
        gs.random.seed(55)

        manifold = hypersphere.Hypersphere(2)
        metric = hypersphere.HypersphereMetric(2)

        x = manifold.random_von_mises_fisher(kappa=100, n_samples=200)

        kmeans = RiemannianKMeans(metric, 1, tol=1e-3)
        kmeans.fit(x)
        center = kmeans.centroids

        mean = FrechetMean(metric=metric)
        mean.fit(x)

        result = metric.dist(center, mean.estimate_)
        expected = 0.
        self.assertAllClose(expected, result, atol=1e-2)

    @geomstats.tests.np_only
    def test_spd_kmeans_fit(self):
        gs.random.seed(0)
        dim = 3
        n_points = 2
        space = spd_matrices.SPDMatrices(dim)
        data = space.random_uniform(n_samples=n_points)
        metric = spd_matrices.SPDMetricAffine(dim)

        kmeans = RiemannianKMeans(metric, n_clusters=1, point_type='matrix')
        kmeans.fit(data)
        result = kmeans.centroids

        mean = FrechetMean(metric=metric, point_type='matrix', max_iter=100)
        mean.fit(data)
        expected = mean.estimate_
        self.assertAllClose(result, expected, atol=1e-2, rtol=1e-2)

    @geomstats.tests.np_and_pytorch_only
    def test_hypersphere_kmeans_predict(self):
        gs.random.seed(1234)
        dim = 2

        manifold = hypersphere.Hypersphere(dim)
        metric = hypersphere.HypersphereMetric(dim)

        x = manifold.random_von_mises_fisher(kappa=100, n_samples=200)

        kmeans = RiemannianKMeans(metric, 5, tol=1e-5)
        kmeans.fit(x, max_iter=100)
        result = kmeans.predict(x)

        centroids = kmeans.centroids
        expected = gs.array(
            [int(metric.closest_neighbor_index(x_i, centroids))
             for x_i in x])
        self.assertAllClose(expected, result)



"""
Helper functions for unit tests.
"""

import geomstats.backend as gs


def to_scalar(expected):
    expected = gs.to_ndarray(expected, to_ndim=1)
    expected = gs.to_ndarray(expected, to_ndim=2, axis=-1)
    return expected


def to_vector(expected):
    expected = gs.to_ndarray(expected, to_ndim=2)
    return expected


def to_matrix(expected):
    expected = gs.to_ndarray(expected, to_ndim=3)
    return expected


def left_log_then_exp_from_identity(metric, point):
    aux = metric.left_log_from_identity(point=point)
    result = metric.left_exp_from_identity(tangent_vec=aux)
    return result


def left_exp_then_log_from_identity(metric, tangent_vec):
    aux = metric.left_exp_from_identity(tangent_vec=tangent_vec)
    result = metric.left_log_from_identity(point=aux)
    return result


def log_then_exp_from_identity(metric, point):
    aux = metric.log_from_identity(point=point)
    result = metric.exp_from_identity(tangent_vec=aux)
    return result


def exp_then_log_from_identity(metric, tangent_vec):
    aux = metric.exp_from_identity(tangent_vec=tangent_vec)
    result = metric.log_from_identity(point=aux)
    return result


def log_then_exp(metric, point, base_point=None):
    aux = metric.log(point=point,
                     base_point=base_point)
    result = metric.exp(tangent_vec=aux,
                        base_point=base_point)
    return result


def exp_then_log(metric, tangent_vec, base_point=None):
    aux = metric.exp(tangent_vec=tangent_vec,
                     base_point=base_point)
    result = metric.log(point=aux,
                        base_point=base_point)
    return result


def group_log_then_exp_from_identity(group, point):
    aux = group.log_from_identity(point=point)
    result = group.exp_from_identity(tangent_vec=aux)
    return result


def group_exp_then_log_from_identity(group, tangent_vec):
    aux = group.exp_from_identity(tangent_vec=tangent_vec)
    result = group.log_from_identity(point=aux)
    return result


def group_log_then_exp(group, point, base_point):
    aux = group.log(
        point=point, base_point=base_point)
    result = group.exp(
        tangent_vec=aux, base_point=base_point)
    return result


def group_exp_then_log(group, tangent_vec, base_point):
    aux = group.exp(
        tangent_vec=tangent_vec, base_point=base_point)
    result = group.log(
        point=aux, base_point=base_point)
    return result



import numpy as np
import torch


def value_and_grad(objective):
    """'Returns a function that returns both value and gradient.

    Suitable for use in scipy.optimize

    Parameters
    ----------
    objective : callable
        Function to compute the gradient. It must be real-valued.

    Returns
    -------
    objective_with_grad : callable
        Function that takes the argument of the objective function as input
        and returns both value and grad at the input.
    '"""
    def objective_with_grad(velocity):
        if isinstance(velocity, np.ndarray):
            velocity = torch.from_numpy(velocity)
        vel = velocity.clone().detach().requires_grad_(True)
        loss = objective(vel)
        loss.backward()
        return loss.detach().numpy(), vel.grad.detach().numpy()
    return objective_with_grad



"""Riemannian and pseudo-Riemannian metrics."""

import autograd

import geomstats.backend as gs
import geomstats.vectorization
from geomstats.geometry.connection import Connection

EPSILON = 1e-4
N_CENTERS = 10
TOLERANCE = 1e-5
N_REPETITIONS = 20
N_MAX_ITERATIONS = 50000
N_STEPS = 10


def loss(y_pred, y_true, metric):
    """Compute loss function between prediction and ground truth.

    Loss function given by a Riemannian metric,
    expressed as the squared geodesic distance between the prediction
    and the ground truth.

    Parameters
    ----------
    y_pred : array-like, shape=[..., dim]
        Prediction.
    y_true : array-like, shape=[..., dim]
        Ground-truth.
    metric : RiemannianMetric
        Metric.

    Returns
    -------
    sq_dist : array-like, shape=[...,]
        Loss, i.e. the squared distance.
    """
    sq_dist = metric.squared_dist(y_pred, y_true)
    return sq_dist


def grad(y_pred, y_true, metric):
    """Closed-form for the gradient of the loss function.

    Parameters
    ----------
    y_pred : array-like, shape=[..., dim]
        Prediction.
    y_true : array-like, shape=[..., dim]
        Ground-truth.
    metric : RiemannianMetric
        Metric.

    Returns
    -------
    loss_grad : array-like, shape=[...,]
        Gradient of the loss.

    """
    tangent_vec = metric.log(base_point=y_pred, point=y_true)
    grad_vec = - 2. * tangent_vec

    inner_prod_mat = metric.inner_product_matrix(base_point=y_pred)
    is_vectorized = inner_prod_mat.ndim == 3
    axes = (0, 2, 1) if is_vectorized else (1, 0)

    loss_grad = gs.einsum(
        '...i,...ij->...i',
        grad_vec,
        gs.transpose(inner_prod_mat, axes=axes))

    return loss_grad


class RiemannianMetric(Connection):
    """Class for Riemannian and pseudo-Riemannian metrics.

    Parameters
    ----------
    dim : int
        Dimension of the manifold.
    signature : tuple
        Signature of the metric.
        Optional, default: None.
    default_point_type : str, {'vector', 'matrix'}
        Point type.
        Optional, default: 'vector'.
    """

    def __init__(self, dim, signature=None, default_point_type='vector'):
        super(RiemannianMetric, self).__init__(
            dim=dim, default_point_type=default_point_type)
        self.signature = signature

    def inner_product_matrix(self, base_point=None):
        """Inner product matrix at the tangent space at a base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.
        """
        raise NotImplementedError(
            'The computation of the inner product matrix'
            ' is not implemented.')

    def inner_product_inverse_matrix(self, base_point=None):
        """Inner product matrix at the tangent space at a base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        mat : array-like, shape=[..., dim, dim]
            Inverse of inner-product matrix.
        """
        metric_matrix = self.inner_product_matrix(base_point)
        cometric_matrix = gs.linalg.inv(metric_matrix)
        return cometric_matrix

    def inner_product_derivative_matrix(self, base_point=None):
        """Compute derivative of the inner prod matrix at base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        mat : array-like, shape=[..., dim, dim]
            Derivative of inverse of inner-product matrix.
        """
        metric_derivative = autograd.jacobian(self.inner_product_matrix)
        return metric_derivative(base_point)

    def christoffels(self, base_point):
        """Compute Christoffel symbols associated with the connection.

        Parameters
        ----------
        base_point: array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        christoffels: array-like, shape=[..., dim, dim, dim]
            Christoffel symbols.
        """
        cometric_mat_at_point = self.inner_product_inverse_matrix(base_point)
        metric_derivative_at_point = self.inner_product_derivative_matrix(
            base_point)
        term_1 = gs.einsum('...im,...mkl->...ikl',
                           cometric_mat_at_point,
                           metric_derivative_at_point)
        term_2 = gs.einsum('...im,...mlk->...ilk',
                           cometric_mat_at_point,
                           metric_derivative_at_point)
        term_3 = - gs.einsum('...im,...klm->...ikl',
                             cometric_mat_at_point,
                             metric_derivative_at_point)

        christoffels = 0.5 * (term_1 + term_2 + term_3)
        return christoffels

    @geomstats.vectorization.decorator(['else', 'vector', 'vector', 'vector'])
    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., dim]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., dim]
            Tangent vector at base point.
        base_point: array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        inner_prod_mat = self.inner_product_matrix(base_point)
        inner_prod_mat = gs.to_ndarray(inner_prod_mat, to_ndim=3)

        aux = gs.einsum('...j,...jk->...k', tangent_vec_a, inner_prod_mat)

        inner_prod = gs.einsum('...k,...k->...', aux, tangent_vec_b)
        inner_prod = gs.to_ndarray(inner_prod, to_ndim=1)
        inner_prod = gs.to_ndarray(inner_prod, to_ndim=2, axis=1)

        return inner_prod

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        sq_norm = self.inner_product(vector, vector, base_point)
        return sq_norm

    def norm(self, vector, base_point=None):
        """Compute norm of a vector.

        Norm of a vector associated to the inner product
        at the tangent space at a base point.

        Note: This only works for positive-definite
        Riemannian metrics and inner products.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        norm : array-like, shape=[...,]
            Norm.
        """
        sq_norm = self.squared_norm(vector, base_point)
        norm = gs.sqrt(sq_norm)
        return norm

    def squared_dist(self, point_a, point_b):
        """Squared geodesic distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            Point.
        point_b : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        sq_dist : array-like, shape=[...,]
            Squared distance.
        """
        log = self.log(point=point_b, base_point=point_a)

        sq_dist = self.squared_norm(vector=log, base_point=point_a)
        return sq_dist

    def dist(self, point_a, point_b):
        """Geodesic distance between two points.

        Note: It only works for positive definite
        Riemannian metrics.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            Point.
        point_b : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        dist : array-like, shape=[...,]
            Distance.
        """
        sq_dist = self.squared_dist(point_a, point_b)
        dist = gs.sqrt(sq_dist)
        return dist

    def dist_pairwise(self, point):
        """Compute the pairwise distance between points.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim]
            Set of points in hyperbolic space.

        Returns
        -------
        dist : array-like, shape=[n_samples, n_samples]
            Pairwise distance matrix between all points.
        """
        pairwise_dist = []

        for i, x in enumerate(point):
            for y in point[i:]:
                pairwise_dist.append(self.dist(x, y))

        pairwise_dist = geomstats.geometry.symmetric_matrices.\
            SymmetricMatrices.from_vector(gs.array(pairwise_dist))

        return pairwise_dist

    def diameter(self, points):
        """Give the distance between two farthest points.

        Distance between the two points that are farthest away from each other
        in points.

        Parameters
        ----------
        points : array-like, shape=[..., dim]
            Points.

        Returns
        -------
        diameter : float
            Distance between two farthest points.
        """
        diameter = 0.0
        n_points = points.shape[0]

        for i in range(n_points - 1):
            dist_to_neighbors = self.dist(points[i, :], points[i + 1:, :])
            dist_to_farthest_neighbor = gs.amax(dist_to_neighbors)
            diameter = gs.maximum(diameter, dist_to_farthest_neighbor)

        return diameter

    def closest_neighbor_index(self, point, neighbors):
        """Closest neighbor of point among neighbors.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point.
        neighbors : array-like, shape=[..., dim]
            Neighbors.

        Returns
        -------
        closest_neighbor_index : int
            Index of closest neighbor.
        """
        dist = self.dist(point, neighbors)
        closest_neighbor_index = gs.argmin(dist)

        return closest_neighbor_index



import logging


logging.basicConfig(format='%(levelname)s: %(message)s')
logging.getLogger().setLevel(logging.INFO)
loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
if loggers and loggers[0].name.startswith('nose2'):
    logging.getLogger().setLevel(logging.WARNING)



"""Manifold module.

In other words, a topological space that locally resembles
Euclidean space near each point.
"""

import geomstats.errors


ATOL = 1e-6


class Manifold:
    r"""Class for manifolds.

    Parameters
    ----------
    dim : int
        Dimension of the manifold.
    default_point_type : str, {\'vector\', \'matrix\'}
        Point type.
        Optional, default: 'vector'.
    default_coords_type : str, {\'intrinsic\', \'extrinsic\', etc}
        Coordinate type.
        Optional, default: 'intrinsic'.
    """

    def __init__(
            self, dim, default_point_type='vector',
            default_coords_type='intrinsic'):
        geomstats.errors.check_integer(dim, 'dim')
        geomstats.errors.check_parameter_accepted_values(
            default_point_type, 'default_point_type', ['vector', 'matrix'])

        self.dim = dim
        self.default_point_type = default_point_type
        self.default_coords_type = default_coords_type

    def belongs(self, point, atol=ATOL):
        """Evaluate if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point to evaluate.
        atol : float
            Absolute tolerance.
            Optional, default: 1e-6.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the manifold.
        """
        raise NotImplementedError('belongs is not implemented.')

    def is_tangent(self, vector, base_point=None, atol=ATOL):
        """Check whether the vector is tangent at base_point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.
            Optional, default: none.
        atol : float
            Absolute tolerance.
            Optional, default: 1e-6.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        raise NotImplementedError(
            'is_tangent is not implemented.')

    def to_tangent(self, vector, base_point=None):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point.
        """
        raise NotImplementedError(
            'to_tangent is not implemented.')

    def regularize(self, point):
        """Regularize a point to the canonical representation for the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        regularized_point : array-like, shape=[..., dim]
            Regularized point.
        """
        regularized_point = point
        return regularized_point



"""Module exposing the `Matrices` and `MatricesMetric` class."""

from functools import reduce

import geomstats.backend as gs
import geomstats.errors
from geomstats.geometry.euclidean import EuclideanMetric
from geomstats.geometry.manifold import Manifold


TOLERANCE = 1e-5


class Matrices(Manifold):
    """Class for the space of matrices (m, n).

    Parameters
    ----------
    m, n : int
        Integers representing the shapes of the matrices: m x n.
    """

    def __init__(self, m, n):
        super(Matrices, self).__init__(dim=m * n)
        geomstats.errors.check_integer(n, 'n')
        geomstats.errors.check_integer(m, 'm')
        self.m = m
        self.n = n
        self.default_point_type = 'matrix'
        self.metric = MatricesMetric(m, n)

    def belongs(self, point):
        """Check if point belongs to the Matrices space.

        Parameters
        ----------
        point : array-like, shape=[..., m, n]
            Point to be checked.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the Matrices space.
        """
        point = gs.to_ndarray(point, to_ndim=3)
        _, mat_dim_1, mat_dim_2 = point.shape
        return mat_dim_1 == self.m & mat_dim_2 == self.n

    @staticmethod
    def equal(mat_a, mat_b, atol=TOLERANCE):
        """Test if matrices a and b are close.

        Parameters
        ----------
        mat_a : array-like, shape=[..., dim1, dim2]
            Matrix.
        mat_b : array-like, shape=[..., dim2, dim3]
            Matrix.
        atol : float
            Tolerance.
            Optional, default: 1e-5.

        Returns
        -------
        eq : array-like, shape=[...,]
            Boolean evaluating if the matrices are close.
        """
        is_vectorized = \
            (gs.ndim(gs.array(mat_a)) == 3) or (gs.ndim(gs.array(mat_b)) == 3)
        axes = (1, 2) if is_vectorized else (0, 1)
        return gs.all(gs.isclose(mat_a, mat_b, atol=atol), axes)

    @staticmethod
    def mul(*args):
        """Compute the product of matrices a1, ..., an.

        Parameters
        ----------
        a1 : array-like, shape=[..., dim_1, dim_2]
            Matrix.
        a2 : array-like, shape=[..., dim_2, dim_3]
            Matrix.
        ...
        an : array-like, shape=[..., dim_n-1, dim_n]
            Matrix.

        Returns
        -------
        mul : array-like, shape=[..., dim_1, dim_n]
            Result of the product of matrices.
        """
        return reduce(gs.matmul, args)

    @classmethod
    def bracket(cls, mat_a, mat_b):
        """Compute the commutator of a and b, i.e. `[a, b] = ab - ba`.

        Parameters
        ----------
        mat_a : array-like, shape=[..., n, n]
            Matrix.
        mat_b : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        mat_c : array-like, shape=[..., n, n]
            Commutator.
        """
        return cls.mul(mat_a, mat_b) - cls.mul(mat_b, mat_a)

    @staticmethod
    def transpose(mat):
        """Return the transpose of matrices.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        transpose : array-like, shape=[..., n, n]
            Transposed matrix.
        """
        is_vectorized = (gs.ndim(gs.array(mat)) == 3)
        axes = (0, 2, 1) if is_vectorized else (1, 0)
        return gs.transpose(mat, axes)

    @classmethod
    def is_symmetric(cls, mat, atol=TOLERANCE):
        """Check if a matrix is symmetric.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.
        atol : float
            Absolute tolerance.
            Optional, default: 1e-5.

        Returns
        -------
        is_sym : array-like, shape=[...,]
            Boolean evaluating if the matrix is symmetric.
        """
        return cls.equal(mat, cls.transpose(mat), atol)

    @classmethod
    def is_skew_symmetric(cls, mat, atol=TOLERANCE):
        """Check if a matrix is skew symmetric.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.
        atol : float
            Absolute tolerance.
            Optional, default: 1e-5.

        Returns
        -------
        is_skew_sym : array-like, shape=[...,]
            Boolean evaluating if the matrix is skew-symmetric.
        """
        return cls.equal(mat, - cls.transpose(mat), atol)

    @classmethod
    def to_symmetric(cls, mat):
        """Make a matrix symmetric, by averaging with its transpose.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        sym : array-like, shape=[..., n, n]
            Symmetric matrix.
        """
        return 1 / 2 * (mat + cls.transpose(mat))

    @classmethod
    def to_skew_symmetric(cls, mat):
        """
        Make a matrix skew-symmetric, by averaging with minus its transpose.

        Parameters
        ----------
        mat : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        skew_sym : array-like, shape=[..., n, n]
            Skew-symmetric matrix.
        """
        return 1 / 2 * (mat - cls.transpose(mat))

    def random_uniform(self, n_samples=1, bound=1.):
        """Sample from a uniform distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Bound.
            Optional, default: 1.

        Returns
        -------
        point : array-like, shape=[m, n] or [n_samples, m, n]
            Sample.
        """
        m, n = self.m, self.n
        size = (n_samples, m, n) if n_samples != 1 else (m, n)
        point = bound * (gs.random.rand(*size) - 0.5)
        return point

    @classmethod
    def congruent(cls, mat_1, mat_2):
        """Compute the congruent action of mat_2 on mat_1.

        This is :math: `mat_2 mat_1 mat_2^T`.

        Parameters
        ----------
        mat_1 : array-like, shape=[..., n, n]
            Matrix.
        mat_2 : array-like, shape=[..., n, n]
            Matrix.

        Returns
        -------
        cong : array-like, shape=[..., n, n]
            Result of the congruent action.
        """
        return cls.mul(mat_2, mat_1, cls.transpose(mat_2))


class MatricesMetric(EuclideanMetric):
    """Euclidean metric on matrices given by Frobenius inner-product.

    Parameters
    ----------
    m, n : int
        Integers representing the shapes of the matrices: m x n.
    """

    def __init__(self, m, n):
        dimension = m * n
        super(MatricesMetric, self).__init__(
            dim=dimension)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute Frobenius inner-product of two tan vecs at `base_point`.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., m, n]
            Tangent vector.
        tangent_vec_b : array-like, shape=[..., m, n]
            Tangent vector.
        base_point : array-like, shape=[..., m, n]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_prod : array-like, shape=[...,]
            Frobenius inner-product of tangent_vec_a and tangent_vec_b.
        """
        inner_prod = gs.einsum(
            '...ij,...ij->...', tangent_vec_a, tangent_vec_b)
        return inner_prod



"""K-means clustering."""

import logging
from random import randint

from sklearn.base import BaseEstimator, ClusterMixin

import geomstats.backend as gs
from geomstats.learning._template import TransformerMixin
from geomstats.learning.frechet_mean import FrechetMean


class RiemannianKMeans(TransformerMixin, ClusterMixin, BaseEstimator):
    """Class for k-means clustering on manifolds.

    K-means algorithm using Riemannian manifolds.

    Parameters
    ----------
    n_clusters : int
        Number of clusters (k value of the k-means).
        Optional, default: 8.
    metric : object of class RiemannianMetric
        The geomstats Riemmanian metric associate to the space used.
    init : str
        How to initialize centroids at the beginning of the algorithm. The
        choice 'random' will select training points as initial centroids
        uniformly at random.
        Optional, default: 'random'.
    tol : float
        Convergence factor. Convergence is achieved when the difference of mean
        distance between two steps is lower than tol.
        Optional, default: 1e-2.
    verbose : int
        If verbose > 0, information will be printed during learning.
        Optional, default: 0.

    Example
    -------
    Available example on the Poincaré Ball and Hypersphere manifolds
    :mod:`examples.plot_kmeans_manifolds`
    """

    def __init__(
            self, metric, n_clusters=8, init='random',
            tol=1e-2, mean_method='default', verbose=0, point_type='vector'):
        self.n_clusters = n_clusters
        self.init = init
        self.metric = metric
        self.tol = tol
        self.verbose = verbose
        self.mean_method = mean_method
        self.point_type = point_type

        self.centroids = None

    def fit(self, X, max_iter=100):
        """Provide clusters centroids and data labels.

        Alternate between computing the mean of each cluster
        and labelling data according to the new positions of the centroids.

        Parameters
        ----------
        X : array-like, shape=[..., n_features]
            Training data, where n_samples is the number of samples and
            n_features is the number of features.
        max_iter : int
            Maximum number of iterations.
            Optional, default: 100.

        Returns
        -------
        self : array-like, shape=[n_clusters,]
            Centroids.
        """
        n_samples = X.shape[0]
        self.centroids = [gs.expand_dims(X[randint(0, n_samples - 1)], 0)
                          for i in range(self.n_clusters)]
        self.centroids = gs.concatenate(self.centroids, axis=0)
        index = 0
        while index < max_iter:
            index += 1

            dists = [gs.to_ndarray(
                     self.metric.dist(self.centroids[i], X), 2, 1)
                     for i in range(self.n_clusters)]
            dists = gs.hstack(dists)
            belongs = gs.argmin(dists, 1)
            old_centroids = gs.copy(self.centroids)
            for i in range(self.n_clusters):
                fold = gs.squeeze(X[belongs == i])

                if len(fold) > 0:

                    mean = FrechetMean(
                        metric=self.metric,
                        method=self.mean_method,
                        max_iter=150,
                        point_type=self.point_type)
                    mean.fit(fold)

                    self.centroids[i] = mean.estimate_
                else:
                    self.centroids[i] = X[randint(0, n_samples - 1)]

            centroids_distances = self.metric.dist(
                old_centroids, self.centroids)

            if gs.mean(centroids_distances) < self.tol:
                if self.verbose > 0:
                    logging.info('Convergence reached after {} '
                                 'iterations'.format(index))

                if self.n_clusters == 1:
                    self.centroids = gs.squeeze(self.centroids, axis=0)

                return gs.copy(self.centroids)

        if index == max_iter:
            logging.warning('K-means maximum number of iterations {} reached. '
                            'The mean may be inaccurate'.format(max_iter))

        if self.n_clusters == 1:
            self.centroids = gs.squeeze(self.centroids, axis=0)
        return gs.copy(self.centroids)

    def predict(self, X):
        """Predict the labels for each data point.

        Label each data point with the cluster having the nearest
        centroid using metric distance.

        Parameters
        ----------
        X : array-like, shape=[..., n_features]
            Input data.

        Returns
        -------
        self : array-like, shape=[...,]
            Array of predicted cluster indices for each sample.
        """
        if self.centroids is None:
            raise RuntimeError('fit needs to be called first.')
        dists = gs.stack(
            [self.metric.dist(centroid, X)
             for centroid in self.centroids],
            axis=1)
        dists = gs.squeeze(dists)

        belongs = gs.argmin(dists, -1)

        return belongs



"""Unit tests for vectorization functions."""

import tests.helper as helper

import geomstats.backend as gs
import geomstats.tests
import geomstats.vectorization


class TestVectorization(geomstats.tests.TestCase):
    def setUp(self):

        class Obj:
            def __init__(self):
                self.default_point_type = 'vector'

        @geomstats.vectorization.decorator(['vector', 'vector'])
        def func(tangent_vec_a, tangent_vec_b):
            result = gs.einsum(
                '...i,...i->...i', tangent_vec_a, tangent_vec_b)
            result = helper.to_vector(result)
            return result

        @geomstats.vectorization.decorator(['vector', 'vector'])
        def func_scalar_output(tangent_vec_a, tangent_vec_b):
            result = gs.einsum(
                '...i,...i->...', tangent_vec_a, tangent_vec_b)
            result = helper.to_scalar(result)
            return result

        @geomstats.vectorization.decorator(['vector', 'vector', 'scalar'])
        def func_scalar_input_output(tangent_vec_a, tangent_vec_b, in_scalar):
            aux = gs.einsum(
                'ni,ni->n', tangent_vec_a, tangent_vec_b)
            result = gs.einsum('n,nk->n', aux, in_scalar)
            result = helper.to_scalar(result)
            return result

        @geomstats.vectorization.decorator(['vector', 'vector', 'scalar'])
        def func_optional_input(tangent_vec_a, tangent_vec_b, in_scalar=None):
            if in_scalar is None:
                in_scalar = gs.array([[1]])
            aux = gs.einsum(
                'ni,ni->n', tangent_vec_a, tangent_vec_b)
            result = gs.einsum('n,nk->n', aux, in_scalar)
            result = helper.to_scalar(result)
            return result

        @geomstats.vectorization.decorator(
            ['else', 'vector', 'else', 'vector', 'scalar'])
        def func_else(else_a, tangent_vec_a, else_b, tangent_vec_b):
            result = (else_a + else_b) * gs.einsum(
                'ni,ni->n', tangent_vec_a, tangent_vec_b)
            result = helper.to_scalar(result)
            return result

        @geomstats.vectorization.decorator(['scalar'])
        def is_scalar_vectorized(scalar):
            is_scalar_vec = gs.ndim(scalar) == 2
            has_dim_1 = gs.shape(scalar)[-1] == 1
            result = is_scalar_vec and has_dim_1
            result = helper.to_scalar(result)
            return result

        @geomstats.vectorization.decorator(['vector'])
        def is_vector_vectorized(vector):
            is_vector_vec = gs.ndim(vector) == 2
            is_vector_vec = helper.to_scalar(is_vector_vec)
            return is_vector_vec

        @geomstats.vectorization.decorator(['matrix'])
        def is_matrix_vectorized(matrix):
            is_matrix_vec = gs.ndim(matrix) == 3
            is_matrix_vec = helper.to_scalar(is_matrix_vec)
            return is_matrix_vec

        @geomstats.vectorization.decorator(
            ['else', 'point', 'point_type'])
        def is_point_type_vector(obj, point, point_type=None):
            is_point_type_vector = point_type == 'vector'
            is_point_type_vector = helper.to_scalar(is_point_type_vector)
            return is_point_type_vector

        @geomstats.vectorization.decorator(
            ['else', 'point', 'point_type'])
        def is_point_type_matrix(obj, point, point_type=None):
            is_point_type_matrix = point_type == 'matrix'
            is_point_type_matrix = helper.to_scalar(is_point_type_matrix)
            return is_point_type_matrix

        @geomstats.vectorization.decorator(
            ['else', 'point', 'point_type'])
        def is_vector_vectorized_with_point_type(obj, point, point_type=None):
            is_vector_vec = gs.ndim(point) == 2
            is_vector_vec = helper.to_scalar(is_vector_vec)
            return is_vector_vec

        @geomstats.vectorization.decorator(
            ['else', 'point', 'point_type'])
        def is_matrix_vectorized_with_point_type(obj, point, point_type=None):
            is_matrix_vec = gs.ndim(point) == 3
            is_matrix_vec = helper.to_scalar(is_matrix_vec)
            return is_matrix_vec

        @geomstats.vectorization.decorator(
            ['else', 'point', 'point_type', 'output_point'])
        def output_1d_vector(obj, point, point_type=None):
            n_samples = point.shape[0]
            result = gs.array([[5.]] * n_samples)
            return result

        self.func = func
        self.func_scalar_output = func_scalar_output
        self.func_scalar_input_output = func_scalar_input_output
        self.func_optional_input = func_optional_input
        self.func_else = func_else
        self.is_scalar_vectorized = is_scalar_vectorized
        self.is_vector_vectorized = is_vector_vectorized
        self.is_matrix_vectorized = is_matrix_vectorized
        self.is_point_type_vector = is_point_type_vector
        self.is_point_type_matrix = is_point_type_matrix
        self.is_vector_vectorized_with_point_type = \
            is_vector_vectorized_with_point_type
        self.is_matrix_vectorized_with_point_type = \
            is_matrix_vectorized_with_point_type
        self.output_1d_vector = output_1d_vector
        self.obj = Obj()

    def test_decorator_with_squeeze_dim0(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        result = self.func(vec_a, vec_b)
        expected = gs.array([0, 2, 0])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_with_squeeze_dim0_with_kwargs(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        result = self.func(tangent_vec_a=vec_a, tangent_vec_b=vec_b)
        expected = gs.array([0, 2, 0])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_without_squeeze_dim0(self):
        vec_a = gs.array([[1, 2, 3]])
        vec_b = gs.array([0, 1, 0])
        result = self.func(vec_a, vec_b)
        expected = gs.array([[0, 2, 0]])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_without_squeeze_dim0_with_kwargs(self):
        vec_a = gs.array([[1, 2, 3]])
        vec_b = gs.array([0, 1, 0])
        result = self.func(tangent_vec_a=vec_a, tangent_vec_b=vec_b)
        expected = gs.array([[0, 2, 0]])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_vectorization(self):
        vec_a = gs.array([[1, 2, 3], [1, 2, 3]])
        vec_b = gs.array([0, 1, 0])
        result = self.func(vec_a, vec_b)
        expected = gs.array([[0, 2, 0], [0, 2, 0]])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_vectorization_with_kwargs(self):
        vec_a = gs.array([[1, 2, 3], [1, 2, 3]])
        vec_b = gs.array([0, 1, 0])
        result = self.func(tangent_vec_a=vec_a, tangent_vec_b=vec_b)
        expected = gs.array([[0, 2, 0], [0, 2, 0]])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_scalar_with_squeeze_dim1(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        result = self.func_scalar_output(vec_a, vec_b)
        expected = 2

        self.assertAllClose(result, expected)

    def test_decorator_scalar_with_squeeze_dim1_with_kwargs(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        result = self.func_scalar_output(
            tangent_vec_a=vec_a, tangent_vec_b=vec_b)
        expected = 2

        self.assertAllClose(result, expected)

    def test_decorator_scalar_without_squeeze_dim1(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        scalar = 4
        result = self.func_scalar_input_output(vec_a, vec_b, scalar)
        expected = 8

        self.assertAllClose(result, expected)

    def test_decorator_scalar_without_squeeze_dim1_with_kwargs(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        scalar = 4
        result = self.func_scalar_input_output(
            tangent_vec_a=vec_a, tangent_vec_b=vec_b, in_scalar=scalar)
        expected = 8

        self.assertAllClose(result, expected)

    def test_decorator_scalar_output_vectorization(self):
        vec_a = gs.array([[1, 2, 3], [1, 2, 3]])
        vec_b = gs.array([0, 1, 0])
        result = self.func_scalar_output(vec_a, vec_b)
        expected = gs.array([2, 2])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_scalar_output_vectorization_with_kwargs(self):
        vec_a = gs.array([[1, 2, 3], [1, 2, 3]])
        vec_b = gs.array([0, 1, 0])
        result = self.func_scalar_output(
            tangent_vec_a=vec_a, tangent_vec_b=vec_b)
        expected = gs.array([2, 2])

        self.assertAllClose(result.shape, expected.shape)
        self.assertAllClose(result, expected)

    def test_decorator_optional_input(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        result = self.func_optional_input(vec_a, vec_b)
        expected = 2

        self.assertAllClose(result, expected)

    def test_decorator_optional_input_with_kwargs(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        result = self.func_optional_input(
            tangent_vec_a=vec_a, tangent_vec_b=vec_b, in_scalar=3)
        expected = 6

        self.assertAllClose(result, expected)

    def test_decorator_optional_input_with_optional_kwargs(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        result = self.func_optional_input(
            tangent_vec_a=vec_a, tangent_vec_b=vec_b)
        expected = 2

        self.assertAllClose(result, expected)

    def test_decorator_else(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        else_a = 1
        else_b = 1
        result = self.func_else(else_a, vec_a, else_b, vec_b)
        expected = 4

        self.assertAllClose(result, expected)

    def test_decorator_else_with_kwargs(self):
        vec_a = gs.array([1, 2, 3])
        vec_b = gs.array([0, 1, 0])
        else_a = 1
        else_b = 1
        result = self.func_else(
            else_a=else_a, tangent_vec_a=vec_a,
            else_b=else_b, tangent_vec_b=vec_b)
        expected = 4

        self.assertAllClose(result, expected)

    def test_is_scalar_vectorized(self):
        scalar = 1.3
        result = self.is_scalar_vectorized(scalar)
        expected = True
        self.assertAllClose(result, expected)

    def test_is_scalar_vectorized_with_kwargs(self):
        scalar = 1.3
        result = self.is_scalar_vectorized(scalar=scalar)
        expected = True
        self.assertAllClose(result, expected)

    def test_is_vector_vectorized(self):
        vector = gs.array([1.3, 3.3])
        result = self.is_vector_vectorized(vector)
        expected = True
        self.assertAllClose(result, expected)

    def test_is_vector_vectorizedi_with_kwargs(self):
        vector = gs.array([1.3, 3.3])
        result = self.is_vector_vectorized(vector=vector)
        expected = True
        self.assertAllClose(result, expected)

    def test_is_matrix_vectorized(self):
        matrix = gs.array([[1.3, 3.3], [1.2, 3.1]])
        result = self.is_matrix_vectorized(matrix)
        expected = True
        self.assertAllClose(result, expected)

    def test_is_matrix_vectorized_with_kwargs(self):
        matrix = gs.array([[1.3, 3.3], [1.2, 3.1]])
        result = self.is_matrix_vectorized(matrix=matrix)
        expected = True
        self.assertAllClose(result, expected)

    def test_vectorize_args(self):
        point_types = ['scalar']
        args = (1.3,)
        result = geomstats.vectorization.vectorize_args(point_types, args)
        expected = (gs.array([[1.3]]),)
        self.assertAllClose(result, expected)

    def test_vectorize_kwargs(self):
        point_types = ['scalar']
        kwargs = {'scalar_name': 1.3}
        result_dict = geomstats.vectorization.vectorize_kwargs(
            point_types, kwargs)
        expected_dict = {'scalar_name': gs.array([[1.3]])}

        keys = expected_dict.keys()
        result = gs.array([result_dict[key] for key in keys])
        expected = gs.array([expected_dict[key] for key in keys])

        self.assertAllClose(result, expected)

    def test_is_point_type_vector(self):
        point = gs.array([1., 2., 3.])
        result = self.is_point_type_vector(
            self.obj, point, point_type='vector')
        expected = True
        self.assertAllClose(result, expected)

    def test_is_point_type_vector_optional(self):
        point = gs.array([1., 2., 3.])
        result = self.is_point_type_vector(
            self.obj, point)
        expected = True
        self.assertAllClose(result, expected)

    def test_is_point_type_matrix(self):
        point = gs.array([[1., 2., 3.], [2., 3., 4.]])
        result = self.is_point_type_matrix(
            self.obj, point, point_type='matrix')
        expected = True
        self.assertAllClose(result, expected)

    def test_is_vector_vectorized_with_point_type(self):
        vector = gs.array([1.3, 3.3])
        result = self.is_vector_vectorized_with_point_type(
            self.obj, point=vector, point_type='vector')
        expected = True
        self.assertAllClose(result, expected)

    def test_is_vector_vectorized_with_optional_point_type(self):
        vector = gs.array([1.3, 3.3])
        result = self.is_vector_vectorized_with_point_type(
            self.obj, point=vector)
        expected = True
        self.assertAllClose(result, expected)

    def test_is_matrix_vectorized_with_point_type(self):
        matrix = gs.array([[1.3, 3.3], [1.2, 3.1]])
        result = self.is_matrix_vectorized_with_point_type(
            self.obj, matrix, point_type='matrix')
        expected = True
        self.assertAllClose(result, expected)

    def test_output_single_1d_vector(self):
        point = gs.array([1., 2., 3.])
        result = self.output_1d_vector(
            self.obj, point)
        expected = gs.array([5.])
        self.assertAllClose(result, expected)

    def test_output_multiple_1d_vector(self):
        point = gs.array([[1., 2., 3.], [2., 3., 4.]])
        result = self.output_1d_vector(
            self.obj, point)
        expected = gs.array([[5.], [5.]])
        self.assertAllClose(result, expected)



"""Unit tests for special euclidean group SE(n).

Note: Only the *canonical* left- and right- invariant
metrics on SE(3) are tested here. Other invariant
metrics are tested with the tests of the invariant_metric
module.
"""

import warnings

import tests.helper as helper

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.invariant_metric import InvariantMetric
from geomstats.geometry.special_euclidean import SpecialEuclidean

# Tolerance for errors on predicted vectors, relative to the *norm*
# of the vector, as opposed to the standard behavior of gs.allclose
# where it is relative to each element of the array

RTOL = 1e-5


class TestSpecialEuclidean3Methods(geomstats.tests.TestCase):
    def setUp(self):
        warnings.simplefilter('ignore', category=ImportWarning)
        gs.random.seed(1234)

        group = SpecialEuclidean(n=3, point_type='vector')

        # Points

        # -- Rotation vectors with angles
        # 0, close to 0, closely lower than pi, pi,
        # between pi and 2pi, closely larger than 2pi, 2pi,
        # and closely larger than 2pi
        with_angle_0 = gs.zeros(6)
        with_angle_close_0 = (1e-10 * gs.array([1., -1., 1., 0., 0., 0.])
                              + gs.array([0., 0., 0., 1., 5., 2]))
        with_angle_close_pi_low = ((gs.pi - 1e-9) / gs.sqrt(2.)
                                   * gs.array([0., 1., -1., 0., 0., 0.])
                                   + gs.array([0., 0., 0., -100., 0., 2.]))
        with_angle_pi = (gs.pi / gs.sqrt(3.)
                         * gs.array([1., 1., -1., 0., 0., 0.])
                         + gs.array([0., 0., 0., -10.2, 0., 2.6]))
        with_angle_close_pi_high = ((gs.pi + 1e-9) / gs.sqrt(3.)
                                    * gs.array([-1., 1., -1., 0., 0., 0.])
                                    + gs.array([0., 0., 0., -100., 0., 2.]))
        with_angle_in_pi_2pi = ((gs.pi + 0.3) / gs.sqrt(5.)
                                * gs.array([-2., 1., 0., 0., 0., 0.])
                                + gs.array([0., 0., 0., -100., 0., 2.]))
        with_angle_close_2pi_low = ((2 * gs.pi - 1e-9) / gs.sqrt(6.)
                                    * gs.array([2., 1., -1., 0., 0., 0.])
                                    + gs.array([0., 0., 0., 8., 555., -2.]))
        with_angle_2pi = (2. * gs.pi / gs.sqrt(3.)
                          * gs.array([1., 1., -1., 0., 0., 0.])
                          + gs.array([0., 0., 0., 1., 8., -10.]))
        with_angle_close_2pi_high = ((2. * gs.pi + 1e-9) / gs.sqrt(2.)
                                     * gs.array([1., 0., -1., 0., 0., 0.])
                                     + gs.array([0., 0., 0., 1., 8., -10.]))

        point_1 = gs.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])
        point_2 = gs.array([0.5, 0., -0.3, 0.4, 5., 60.])

        translation_large = gs.array([0., 0., 0., 0.4, 0.5, 0.6])
        translation_small = gs.array([0., 0., 0., 0.5, 0.6, 0.7])
        rot_with_parallel_trans = gs.array([gs.pi / 3., 0., 0.,
                                           1., 0., 0.])

        elements_all = {
            'with_angle_0': with_angle_0,
            'with_angle_close_0': with_angle_close_0,
            'with_angle_close_pi_low': with_angle_close_pi_low,
            'with_angle_pi': with_angle_pi,
            'with_angle_close_pi_high': with_angle_close_pi_high,
            'with_angle_in_pi_2pi': with_angle_in_pi_2pi,
            'with_angle_close_2pi_low': with_angle_close_2pi_low,
            'with_angle_2pi': with_angle_2pi,
            'with_angle_close_2pi_high': with_angle_close_2pi_high,
            'translation_large': translation_large,
            'translation_small': translation_small,
            'point_1': point_1,
            'point_2': point_2,
            'rot_with_parallel_trans': rot_with_parallel_trans}
        elements = elements_all
        if geomstats.tests.tf_backend():
            # Tf is extremely slow
            elements = {
                'point_1': point_1,
                'point_2': point_2}

        elements_matrices_all = {
            key: group.matrix_from_vector(elements_all[key]) for key in
            elements_all}
        elements_matrices = elements_matrices_all

        # Metrics - only diagonals
        diag_mat_at_identity = gs.eye(6) * gs.array([2., 2., 2., 3., 3., 3.])

        left_diag_metric = InvariantMetric(
            group=group,
            inner_product_mat_at_identity=diag_mat_at_identity,
            left_or_right='left')
        right_diag_metric = InvariantMetric(
            group=group,
            inner_product_mat_at_identity=diag_mat_at_identity,
            left_or_right='right')

        # mat_at_identity = 7 * gs.eye(group.dim)

        # left_metric = InvariantMetric(
        #            group=group,
        #            inner_product_mat_at_identity=mat_at_identity,
        #            left_or_right='left')
        # right_metric = InvariantMetric(
        #            group=group,
        #            inner_product_mat_at_identity=mat_at_identity,
        #            left_or_right='right')

        metrics_all = {
            'left_canonical': group.left_canonical_metric,
            'right_canonical': group.right_canonical_metric,
            'left_diag': left_diag_metric,
            'right_diag': right_diag_metric}
        # FIXME:
        # 'left': left_metric,
        # 'right': right_metric}
        metrics = metrics_all
        if geomstats.tests.tf_backend():
            metrics = {'left_diag': left_diag_metric}

        self.group = group
        self.metrics_all = metrics_all
        self.metrics = metrics
        self.elements_all = elements_all
        self.elements = elements
        self.elements_matrices_all = elements_matrices_all
        self.elements_matrices = elements_matrices
        self.angles_close_to_pi_all = [
            'with_angle_close_pi_low',
            'with_angle_pi',
            'with_angle_close_pi_high']
        self.angles_close_to_pi = self.angles_close_to_pi_all
        if geomstats.tests.tf_backend():
            self.angles_close_to_pi = ['with_angle_close_pi_low']

        self.n_samples = 4

    def test_random_and_belongs(self):
        """Checks random_uniform and belongs

        Test that the random uniform method samples
        on the special euclidean group.
        """
        base_point = self.group.random_uniform()
        result = self.group.belongs(base_point)
        expected = True
        self.assertAllClose(result, expected)

    def test_random_and_belongs_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_uniform(n_samples=n_samples)
        result = self.group.belongs(points)
        expected = gs.array([True] * n_samples)
        self.assertAllClose(result, expected)

    def test_regularize(self):
        point = self.elements_all['with_angle_0']
        result = self.group.regularize(point)
        expected = point
        self.assertAllClose(result, expected)

        less_than_pi = ['with_angle_close_0',
                        'with_angle_close_pi_low']
        for angle_type in less_than_pi:
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)
            expected = point
            self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Note: by default, the rotation vector is inverted by
            # the function regularize when the angle of the rotation is pi.
            angle_type = 'with_angle_pi'
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)

            expected = point

            self.assertAllClose(result, expected)

            angle_type = 'with_angle_close_pi_high'
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)
            expected_rot = gs.concatenate(
                [point[:3] / gs.linalg.norm(point[:3]) * gs.pi,
                 gs.zeros(3)], axis=0)
            expected_trans = gs.concatenate(
                [gs.zeros(3), point[3:6]], axis=0)
            expected = expected_rot + expected_trans
            self.assertAllClose(result, expected)

            in_pi_2pi = ['with_angle_in_pi_2pi',
                         'with_angle_close_2pi_low']

            for angle_type in in_pi_2pi:
                point = self.elements_all[angle_type]
                angle = gs.linalg.norm(point[:3])
                new_angle = gs.pi - (angle - gs.pi)

                result = self.group.regularize(point)
                expected_rot = gs.concatenate(
                    [- new_angle * (point[:3] / angle),
                     gs.zeros(3)], axis=0)
                expected_trans = gs.concatenate(
                    [gs.zeros(3),
                     point[3:6]], axis=0)
                expected = expected_rot + expected_trans

                self.assertAllClose(result, expected)

            angle_type = 'with_angle_2pi'
            point = self.elements_all[angle_type]
            result = self.group.regularize(point)
            expected = gs.concatenate([gs.zeros(3), point[3:6]], axis=0)
            self.assertAllClose(result, expected)

            angle_type = 'with_angle_close_2pi_high'
            point = self.elements_all[angle_type]
            angle = gs.linalg.norm(point[:3])
            new_angle = angle - 2 * gs.pi

            result = self.group.regularize(point)

            expected_rot = gs.concatenate(
                [new_angle * point[:3] / angle,
                 gs.zeros(3)], axis=0)
            expected_trans = gs.concatenate(
                [gs.zeros(3), point[3:6]], axis=0)
            expected = expected_rot + expected_trans
            self.assertAllClose(result, expected)

    def test_regularize_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_uniform(n_samples=n_samples)
        regularized_points = self.group.regularize(points)

        self.assertAllClose(
            gs.shape(regularized_points),
            (n_samples, *self.group.get_point_type_shape()))

    def test_compose(self):
        # Composition by identity, on the right
        # Expect the original transformation
        point = self.elements_all['point_1']
        result = self.group.compose(point,
                                    self.group.identity)
        expected = point
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Composition by identity, on the left
            # Expect the original transformation
            result = self.group.compose(self.group.identity,
                                        point)
            expected = point
            self.assertAllClose(result, expected)

            # Composition of translations (no rotational part)
            # Expect the sum of the translations
            result = self.group.compose(self.elements_all['translation_small'],
                                        self.elements_all['translation_large'])
            expected = (self.elements_all['translation_small']
                        + self.elements_all['translation_large'])
            self.assertAllClose(result, expected)

    def test_compose_and_inverse(self):
        point = self.elements_all['point_1']
        inv_point = self.group.inverse(point)
        # Compose transformation by its inverse on the right
        # Expect the group identity
        result = self.group.compose(point, inv_point)
        expected = self.group.identity
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Compose transformation by its inverse on the left
            # Expect the group identity
            result = self.group.compose(inv_point, point)
            expected = self.group.identity
            self.assertAllClose(result, expected)

    def test_compose_vectorization(self):
        n_samples = self.n_samples
        n_points_a = self.group.random_uniform(n_samples=n_samples)
        n_points_b = self.group.random_uniform(n_samples=n_samples)
        one_point = self.group.random_uniform(n_samples=1)

        result = self.group.compose(one_point,
                                    n_points_a)
        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape()))

        result = self.group.compose(n_points_a,
                                    one_point)

        if not geomstats.tests.tf_backend():
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            result = self.group.compose(n_points_a,
                                        n_points_b)
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

    def test_inverse_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_uniform(n_samples=n_samples)
        result = self.group.inverse(points)
        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape()))

    def test_left_jacobian_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_uniform(n_samples=n_samples)
        result = self.group.jacobian_translation(
            point=points, left_or_right='left')
        self.assertAllClose(
            gs.shape(result),
            (n_samples, *self.group.get_point_type_shape(),
             *self.group.get_point_type_shape()))

    def test_exp_from_identity_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            tangent_vecs = self.group.random_uniform(n_samples=n_samples)
            result = metric.exp_from_identity(tangent_vecs)

            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            if geomstats.tests.tf_backend():
                break

    def test_log_from_identity_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            points = self.group.random_uniform(n_samples=n_samples)
            result = metric.log_from_identity(points)

            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            if geomstats.tests.tf_backend():
                break

    def test_exp_vectorization(self):
        n_samples = self.n_samples

        for metric_type in self.metrics:
            metric = self.metrics[metric_type]

            one_tangent_vec = self.group.random_uniform(n_samples=1)
            one_base_point = self.group.random_uniform(n_samples=1)
            n_tangent_vec = self.group.random_uniform(n_samples=n_samples)
            n_base_point = self.group.random_uniform(n_samples=n_samples)

            # Test with the 1 base point, and n tangent vecs
            result = metric.exp(n_tangent_vec, one_base_point)
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            if geomstats.tests.tf_backend():
                break

            # Test with the several base point, and one tangent vec
            result = metric.exp(one_tangent_vec, n_base_point)
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            # Test with the same number n of base point and n tangent vec
            result = metric.exp(n_tangent_vec, n_base_point)
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

    def test_log_vectorization(self):
        n_samples = self.n_samples
        for metric_type in self.metrics:
            metric = self.metrics[metric_type]

            one_point = self.group.random_uniform(n_samples=1)
            one_base_point = self.group.random_uniform(n_samples=1)
            n_point = self.group.random_uniform(n_samples=n_samples)
            n_base_point = self.group.random_uniform(n_samples=n_samples)

            # Test with the 1 base point, and several different points
            result = metric.log(n_point, one_base_point)
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            if geomstats.tests.tf_backend():
                break

            # Test with the several base point, and 1 point
            result = metric.log(one_point, n_base_point)
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            # Test with the same number n of base point and point
            result = metric.log(n_point, n_base_point)
            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

    @geomstats.tests.np_only
    def test_group_exp_from_identity_vectorization(self):
        n_samples = self.n_samples
        tangent_vecs = self.group.random_uniform(n_samples=n_samples)
        result = self.group.exp_from_identity(tangent_vecs)

        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape()))

    @geomstats.tests.np_only
    def test_group_log_from_identity_vectorization(self):
        n_samples = self.n_samples
        points = self.group.random_uniform(n_samples=n_samples)
        result = self.group.log_from_identity(points)

        self.assertAllClose(
            gs.shape(result),
            (n_samples, *self.group.get_point_type_shape()))

    @geomstats.tests.np_only
    def test_group_exp_vectorization(self):
        n_samples = self.n_samples
        # Test with the 1 base_point, and several different tangent_vecs
        tangent_vecs = self.group.random_uniform(n_samples=n_samples)
        base_point = self.group.random_uniform(n_samples=1)
        result = self.group.exp(tangent_vecs, base_point)

        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape()))

        if not geomstats.tests.tf_backend():
            # Test with the same number of base_points and tangent_vecs
            tangent_vecs = self.group.random_uniform(n_samples=n_samples)
            base_points = self.group.random_uniform(n_samples=n_samples)
            result = self.group.exp(tangent_vecs, base_points)

            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            # Test with the several base_points, and 1 tangent_vec
            tangent_vec = self.group.random_uniform(n_samples=1)
            base_points = self.group.random_uniform(n_samples=n_samples)
            result = self.group.exp(tangent_vec, base_points)

            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

    @geomstats.tests.np_only
    def test_group_log_vectorization(self):
        n_samples = self.n_samples
        # Test with the 1 base point, and several different points
        points = self.group.random_uniform(n_samples=n_samples)
        base_point = self.group.random_uniform(n_samples=1)
        result = self.group.log(points, base_point)

        self.assertAllClose(
            gs.shape(result), (n_samples, *self.group.get_point_type_shape()))

        if not geomstats.tests.tf_backend():

            # Test with the same number of base points and points
            points = self.group.random_uniform(n_samples=n_samples)
            base_points = self.group.random_uniform(n_samples=n_samples)
            result = self.group.log(points, base_points)

            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

            # Test with the several base points, and 1 point
            point = self.group.random_uniform(n_samples=1)
            base_points = self.group.random_uniform(n_samples=n_samples)
            result = self.group.log(point, base_points)

            self.assertAllClose(
                gs.shape(result),
                (n_samples, *self.group.get_point_type_shape()))

    @geomstats.tests.np_only
    def test_group_exp_from_identity(self):
        # Group exponential of a translation (no rotational part)
        # Expect the original translation
        tangent_vec = self.elements_all['translation_small']
        result = self.group.exp(
            base_point=self.group.identity, tangent_vec=tangent_vec)
        expected = tangent_vec
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Group exponential of a transformation
            # where translation is parallel to rotation axis
            # Expect the original transformation
            tangent_vec = self.elements_all['rot_with_parallel_trans']
            result = self.group.exp(
                base_point=self.group.identity,
                tangent_vec=tangent_vec)
            expected = tangent_vec
            self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_group_log_from_identity(self):
        # Group logarithm of a translation (no rotational part)
        # Expect the original translation
        point = self.elements_all['translation_small']
        result = self.group.log(
            base_point=self.group.identity, point=point)
        expected = point
        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Group logarithm of a transformation
            # where translation is parallel to rotation axis
            # Expect the original transformation
            point = self.elements_all['rot_with_parallel_trans']
            result = self.group.log(
                base_point=self.group.identity, point=point)
            expected = point
            self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_group_log_then_exp_from_identity(self):
        """
        Test that the group exponential from the identity
        and the group logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        for element_type in self.elements:
            if element_type in self.angles_close_to_pi:
                continue
            point = self.elements[element_type]
            result = helper.group_log_then_exp_from_identity(
                group=self.group, point=point)
            expected = self.group.regularize(point)
            self.assertAllClose(result, expected, atol=1e-3)

            if geomstats.tests.tf_backend():
                break

    @geomstats.tests.np_only
    def test_group_log_then_exp_from_identity_with_angles_close_to_pi(self):
        """
        Test that the group exponential from the identity
        and the group logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        for element_type in angle_types:
            point = self.elements_all[element_type]
            result = helper.group_log_then_exp_from_identity(
                group=self.group, point=point)
            expected = self.group.regularize(point)

            inv_expected = gs.concatenate(
                [- expected[:3], expected[3:6]])

            self.assertTrue(
                gs.allclose(result, expected, atol=1e-4)
                or gs.allclose(result, inv_expected, atol=1e-4))

            if geomstats.tests.tf_backend():
                break

    @geomstats.tests.np_only
    def test_group_exp(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        # Tangent vector is a translation (no infinitesimal rotational part)
        # Expect the sum of the translation
        # with the translation of the reference point
        result = self.group.exp(
            base_point=self.elements_all['translation_small'],
            tangent_vec=self.elements_all['translation_large'])
        expected = (self.elements_all['translation_small']
                    + self.elements_all['translation_large'])
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_group_log(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        # Point is a translation (no rotational part)
        # Expect the difference of the translation
        # by the translation of the reference point
        result = self.group.log(
            base_point=self.elements_all['translation_small'],
            point=self.elements_all['translation_large'])
        expected = (self.elements_all['translation_large']
                    - self.elements_all['translation_small'])

        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_group_log_then_exp(self):
        """
        Test that the group exponential
        and the group logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for base_point in self.elements.values():
            for element_type in self.elements:
                if element_type in self.angles_close_to_pi:
                    continue
                point = self.elements[element_type]

                result = helper.group_log_then_exp(group=self.group,
                                                   point=point,
                                                   base_point=base_point)
                expected = self.group.regularize(point)
                self.assertAllClose(result, expected, rtol=1e-4, atol=1e-4)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_only
    def test_group_exp_then_log(self):
        """
        Test that the group exponential
        and the group logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for base_point_type in self.elements:
            base_point = self.elements[base_point_type]
            for element_type in self.elements:
                if element_type in self.angles_close_to_pi:
                    continue
                tangent_vec = self.elements[element_type]
                result = helper.group_exp_then_log(
                    group=self.group,
                    tangent_vec=tangent_vec,
                    base_point=base_point)
                metric = self.metrics_all['left_canonical']
                expected = self.group.regularize_tangent_vec(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    metric=metric)
                self.assertAllClose(result, expected, rtol=1e-4, atol=1e-4)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_only
    def test_exp_from_identity_left(self):
        # Riemannian left-invariant metric given by
        # the canonical inner product on the lie algebra
        # Expect the identity function
        # because we use the Riemannian left logarithm with canonical
        # inner product to parameterize the transformations
        metric = self.metrics_all['left_canonical']
        # General case
        tangent_rot_vec = gs.array([1., 1., 1.])  # NB: Regularized
        tangent_translation = gs.array([1., 0., -3.])
        tangent_vec = gs.concatenate(
            [tangent_rot_vec, tangent_translation],
            axis=0)
        result = metric.exp_from_identity(tangent_vec)
        expected = tangent_vec

        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_log_from_identity_left(self):
        # Riemannian left-invariant metric given by
        # the canonical inner product on the lie algebra
        # Expect the identity function
        # because we use the Riemannian left logarithm with canonical
        # inner product to parameterize the transformations

        metric = self.metrics_all['left_canonical']
        # General case
        rot_vec = gs.array([0.1, 1, 0.9])  # NB: Regularized
        translation = gs.array([1., -19., -3.])
        transfo = gs.concatenate(
            [rot_vec, translation], axis=0)

        expected = transfo
        result = metric.log_from_identity(transfo)

        self.assertAllClose(result, expected)

        if not geomstats.tests.tf_backend():
            # Edge case: angle < epsilon, where angle = norm(rot_vec)
            rot_vec = gs.array([1e-8, 0., 1e-9])  # NB: Regularized
            translation = gs.array([10000., -5.9, -93])
            transfo = gs.concatenate(
                [rot_vec, translation], axis=0)

            expected = transfo
            result = metric.log_from_identity(transfo)

            self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_exp_then_log_from_identity_left(self):
        """
        Test that the Riemannian left exponential from the identity
        and the Riemannian left logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        # Canonical inner product on the lie algebra

        for metric in [self.metrics_all['left_canonical'],
                       self.metrics_all['left_diag']]:
            for angle_type in self.elements:
                if angle_type in self.angles_close_to_pi:
                    continue
                tangent_vec = self.elements[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec)
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec, metric=metric)
                self.assertAllClose(result, expected)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_only
    def test_exp_then_log_from_identity_left_with_angles_close_to_pi(self):
        """
        Test that the Riemannian left exponential from the identity
        and the Riemannian left logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [self.metrics_all['left_canonical'],
                       self.metrics_all['left_diag']]:
            for angle_type in angle_types:
                tangent_vec = self.elements_all[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec)
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec=tangent_vec, metric=metric)
                inv_expected = gs.concatenate(
                    [- expected[:3], expected[3:6]])

                self.assertTrue(
                    gs.allclose(result, expected)
                    or gs.allclose(result, inv_expected))

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_only
    def test_exp_then_log_from_identity_right(self):
        """
        Test that the Riemannian right exponential from the identity
        and the Riemannian right logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        # Canonical inner product on the lie algebra
        for metric in [self.metrics_all['right_canonical'],
                       self.metrics_all['right_diag']]:
            for angle_type in self.elements:
                if angle_type in self.angles_close_to_pi:
                    continue
                tangent_vec = self.elements[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec)
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec=tangent_vec, metric=metric)

                self.assertAllClose(result, expected, atol=1e-4)

                if geomstats.tests.tf_backend():
                    break

    @geomstats.tests.np_only
    def test_exp_then_log_from_identity_right_with_angles_close_to_pi(self):
        """
        Test that the Riemannian right exponential from the identity
        and the Riemannian right logarithm from the identity
        are inverse.
        Expect their composition to give the identity function.
        """
        if geomstats.tests.np_backend():
            atol = 1e-5
        else:
            atol = 5e-5
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [self.metrics_all['right_canonical'],
                       self.metrics_all['right_diag']]:
            for angle_type in angle_types:
                tangent_vec = self.elements_all[angle_type]
                result = helper.exp_then_log_from_identity(
                    metric=metric, tangent_vec=tangent_vec)
                expected = self.group.regularize_tangent_vec_at_identity(
                    tangent_vec=tangent_vec, metric=metric)
                inv_expected = gs.concatenate(
                    [- expected[:3], expected[3:6]])
                self.assertTrue(
                    gs.allclose(result, expected, atol=atol)
                    or gs.allclose(result, inv_expected, atol=atol))

    @geomstats.tests.np_only
    def test_exp_left(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        metric = self.metrics_all['left_canonical']
        rot_vec_base_point = gs.array([0., 0., 0.])
        translation_base_point = gs.array([4., -1., 10000.])
        transfo_base_point = gs.concatenate(
            [rot_vec_base_point, translation_base_point], axis=0)

        # Tangent vector is a translation (no infinitesimal rotational part)
        # Expect the sum of the translation
        # with the translation of the reference point
        rot_vec = gs.array([0., 0., 0.])
        translation = gs.array([1., 0., -3.])
        tangent_vec = gs.concatenate([rot_vec, translation], axis=0)

        result = metric.exp(base_point=transfo_base_point,
                            tangent_vec=tangent_vec)
        expected = gs.concatenate(
            [gs.array([0., 0., 0.]), gs.array([5., -1., 9997.])],
            axis=0)
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_log_left(self):
        # Reference point is a translation (no rotational part)
        # so that the jacobian of the left-translation of the Lie group
        # is the 6x6 identity matrix
        metric = self.metrics_all['left_canonical']
        rot_vec_base_point = gs.array([0., 0., 0.])
        translation_base_point = gs.array([4., 0., 0.])
        transfo_base_point = gs.concatenate(
            [rot_vec_base_point, translation_base_point], axis=0)

        # Point is a translation (no rotational part)
        # Expect the difference of the translation
        # by the translation of the reference point
        rot_vec = gs.array([0., 0., 0.])
        translation = gs.array([-1., -1., -1.2])
        point = gs.concatenate(
            [rot_vec, translation], axis=0)

        expected = gs.concatenate(
            [gs.array([0., 0., 0.]), gs.array([-5., -1., -1.2])],
            axis=0)

        result = metric.log(base_point=transfo_base_point,
                            point=point)

        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_log_then_exp_left(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for metric in [self.metrics_all['left_canonical'],
                       self.metrics_all['left_diag']]:
            for base_point_type in self.elements:
                base_point = self.elements[base_point_type]
                for element_type in self.elements:
                    if element_type in self.angles_close_to_pi:
                        continue
                    point = self.elements[element_type]
                    result = helper.log_then_exp(
                        metric=metric,
                        point=point,
                        base_point=base_point)

                    expected = self.group.regularize(point)

                    self.assertAllClose(result, expected, atol=1e-4)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_only
    def test_log_then_exp_left_with_angles_close_to_pi(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [self.metrics_all['left_canonical'],
                       self.metrics_all['left_diag']]:
            for base_point in self.elements.values():
                for element_type in angle_types:
                    point = self.elements_all[element_type]
                    result = helper.log_then_exp(
                        metric=metric,
                        point=point,
                        base_point=base_point)

                    expected = self.group.regularize(point)

                    inv_expected = gs.concatenate(
                        [- expected[:3], expected[3:6]])

                    self.assertTrue(
                        gs.allclose(result, expected, atol=1e-6)
                        or gs.allclose(result, inv_expected, atol=1e-6))

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_only
    def test_exp_then_log_left(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for metric in [self.metrics_all['left_canonical'],
                       self.metrics_all['left_diag']]:
            for base_point in self.elements.values():
                for element_type in self.elements:
                    if element_type in self.angles_close_to_pi:
                        continue
                    tangent_vec = self.elements[element_type]
                    result = helper.exp_then_log(
                        metric=metric,
                        tangent_vec=tangent_vec,
                        base_point=base_point)

                    expected = self.group.regularize_tangent_vec(
                        tangent_vec=tangent_vec,
                        base_point=base_point,
                        metric=metric)
                    norm = gs.linalg.norm(expected)
                    atol = RTOL
                    if norm != 0:
                        atol = RTOL * norm
                    self.assertAllClose(result, expected, atol=atol)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_only
    def test_exp_then_log_left_with_angles_close_to_pi(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [self.metrics_all['left_canonical'],
                       self.metrics_all['left_diag']]:
            for base_point in self.elements.values():
                for element_type in angle_types:
                    tangent_vec = self.elements_all[element_type]
                    result = helper.exp_then_log(
                        metric=metric,
                        tangent_vec=tangent_vec,
                        base_point=base_point)

                    expected = self.group.regularize_tangent_vec(
                        tangent_vec=tangent_vec,
                        base_point=base_point,
                        metric=metric)

                    inv_expected = gs.concatenate(
                        [- expected[:3], expected[3:6]])

                    self.assertTrue(
                        gs.allclose(result, expected, atol=1e-3)
                        or gs.allclose(result, inv_expected, atol=1e-3))

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_only
    def test_log_then_exp_right(self):
        """
        Test that the Riemannian right exponential and the
        Riemannian right logarithm are inverse.
        Expect their composition to give the identity function.
        """
        for metric in [self.metrics_all['right_canonical'],
                       self.metrics_all['right_diag']]:
            for base_point in self.elements.values():
                for element_type in self.elements:
                    if element_type in self.angles_close_to_pi:
                        continue
                    point = self.elements[element_type]
                    result = helper.log_then_exp(
                        metric=metric,
                        point=point,
                        base_point=base_point)

                    expected = self.group.regularize(point)
                    norm = gs.linalg.norm(expected)
                    atol = RTOL
                    if norm != 0:
                        atol = RTOL * norm
                    self.assertAllClose(result, expected, atol=atol)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_only
    def test_log_then_exp_right_with_angles_close_to_pi(self):
        """
        Test that the Riemannian right exponential and the
        Riemannian right logarithm are inverse.
        Expect their composition to give the identity function.
        """
        angle_types = self.angles_close_to_pi
        # Canonical inner product on the lie algebra
        for metric in [self.metrics_all['right_canonical'],
                       self.metrics_all['right_diag']]:
            for base_point in self.elements.values():
                for element_type in angle_types:
                    point = self.elements_all[element_type]
                    result = helper.log_then_exp(
                        metric=metric,
                        point=point,
                        base_point=base_point)

                    expected = self.group.regularize(point)

                    inv_expected = gs.concatenate(
                        [- expected[:3], expected[3:6]])

                    norm = gs.linalg.norm(expected)
                    atol = RTOL
                    if norm != 0:
                        atol = RTOL * norm

                    self.assertTrue(
                        gs.allclose(result, expected, atol=atol)
                        or gs.allclose(result, inv_expected, atol=atol))

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_only
    def test_exp_then_log_right(self):
        """
        Test that the Riemannian left exponential and the
        Riemannian left logarithm are inverse.
        Expect their composition to give the identity function.
        """
        # FIXME
        # for metric in [self.metrics_all['right_canonical'],
        #                self.metrics_all['right_diag']]:
        #     for base_point_type in self.elements:
        #         base_point = self.elements[base_point_type]
        #         for element_type in self.elements:
        #             if element_type in self.angles_close_to_pi:
        #                 continue
        #             tangent_vec = self.elements[element_type]
        #             result = helper.exp_then_log(
        #                                         metric=metric,
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point)

        #             expected = self.group.regularize_tangent_vec(
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point,
        #                                         metric=metric)

        #             if geomstats.tests.tf_backend():
        #                 break

    @geomstats.tests.np_only
    def test_exp_then_log_right_with_angles_close_to_pi(self):
        """
        Test that the Riemannian right exponential and the
        Riemannian right logarithm are inverse.
        Expect their composition to give the identity function.
        """
        # FIXME
        # angle_types = self.angles_close_to_pi
        # # Canonical inner product on the lie algebra
        # for metric in [self.metrics_all['right_canonical'],
        #                self.metrics_all['right_diag']]:
        #     for base_point in self.elements.values():
        #         for element_type in angle_types:
        #             tangent_vec = self.elements_all[element_type]
        #             result = helper.exp_then_log(
        #                                         metric=metric,
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point)

        #             expected = self.group.regularize_tangent_vec(
        #                                         tangent_vec=tangent_vec,
        #                                         base_point=base_point,
        #                                         metric=metric)

        #             inv_expected = gs.concatenate(
        #                 [- expected[:, :3], expected[:, 3:6]],
        #                 axis=1)
        #             norm = gs.linalg.norm(expected)
        #             atol = RTOL
        #             if norm != 0:
        #                 atol = RTOL * norm

        #             if geomstats.tests.tf_backend():
        #                 break

    def test_inner_product_at_identity_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            one_vector_a = self.group.random_uniform(n_samples=1)
            one_vector_b = self.group.random_uniform(n_samples=1)
            n_vector_a = self.group.random_uniform(n_samples=n_samples)
            n_vector_b = self.group.random_uniform(n_samples=n_samples)

            result = metric.inner_product(one_vector_a, n_vector_b)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            result = metric.inner_product(n_vector_a, one_vector_b)
            self.assertAllClose(gs.shape(result), (n_samples,))

            result = metric.inner_product(n_vector_a, n_vector_b)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_inner_product_one_base_point_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            one_base_point = self.group.random_uniform(n_samples=1)

            one_vector_a = self.group.random_uniform(n_samples=1)
            one_vector_b = self.group.random_uniform(n_samples=1)
            n_vector_a = self.group.random_uniform(n_samples=n_samples)
            n_vector_b = self.group.random_uniform(n_samples=n_samples)

            result = metric.inner_product(one_vector_a, n_vector_b,
                                          one_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            result = metric.inner_product(n_vector_a, one_vector_b,
                                          one_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            result = metric.inner_product(n_vector_a, n_vector_b,
                                          one_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_inner_product_n_base_point_vectorization(self):
        n_samples = self.n_samples
        for metric in self.metrics.values():
            n_base_point = self.group.random_uniform(n_samples=n_samples)

            one_vector_a = self.group.random_uniform(n_samples=1)
            one_vector_b = self.group.random_uniform(n_samples=1)
            n_vector_a = self.group.random_uniform(n_samples=n_samples)
            n_vector_b = self.group.random_uniform(n_samples=n_samples)

            result = metric.inner_product(one_vector_a, n_vector_b,
                                          n_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            result = metric.inner_product(n_vector_a, one_vector_b,
                                          n_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

            result = metric.inner_product(n_vector_a, n_vector_b,
                                          n_base_point)
            self.assertAllClose(gs.shape(result), (n_samples,))

    @geomstats.tests.np_only
    def test_squared_dist_is_symmetric(self):
        for metric in self.metrics.values():
            for point_a in self.elements.values():
                for point_b in self.elements.values():
                    point_a = self.group.regularize(point_a)
                    point_b = self.group.regularize(point_b)

                    sq_dist_a_b = metric.squared_dist(point_a, point_b)
                    sq_dist_b_a = metric.squared_dist(point_b, point_a)

                    self.assertAllClose(sq_dist_a_b, sq_dist_b_a)

                    if geomstats.tests.tf_backend():
                        break

    @geomstats.tests.np_only
    def test_dist_is_symmetric(self):
        for metric in self.metrics.values():
            for point_a in self.elements.values():
                for point_b in self.elements.values():
                    point_a = self.group.regularize(point_a)
                    point_b = self.group.regularize(point_b)

                    dist_a_b = metric.dist(point_a, point_b)
                    dist_b_a = metric.dist(point_b, point_a)
                    self.assertAllClose(dist_a_b, dist_b_a)

                    if geomstats.tests.tf_backend():
                        break

    def test_squared_dist_vectorization(self):
        n_samples = self.n_samples
        for metric_type in self.metrics:
            metric = self.metrics[metric_type]
            point_id = self.group.identity

            one_point_1 = self.group.random_uniform(n_samples=1)
            one_point_2 = self.group.random_uniform(n_samples=1)
            one_point_1 = self.group.regularize(one_point_1)
            one_point_2 = self.group.regularize(one_point_2)

            n_point_1 = self.group.random_uniform(n_samples=n_samples)
            n_point_2 = self.group.random_uniform(n_samples=n_samples)
            n_point_1 = self.group.regularize(n_point_1)
            n_point_2 = self.group.regularize(n_point_2)

            # Identity and n points 2
            result = metric.squared_dist(point_id, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break

            # n points 1 and identity
            result = metric.squared_dist(n_point_1, point_id)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # one point 1 and n points 2
            result = metric.squared_dist(one_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and one point 2
            result = metric.squared_dist(n_point_1, one_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and n points 2
            result = metric.squared_dist(n_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_dist_vectorization(self):
        n_samples = self.n_samples
        for metric_type in self.metrics:
            metric = self.metrics[metric_type]
            point_id = self.group.identity

            one_point_1 = self.group.random_uniform(n_samples=1)
            one_point_2 = self.group.random_uniform(n_samples=1)
            one_point_1 = self.group.regularize(one_point_1)
            one_point_2 = self.group.regularize(one_point_2)

            n_point_1 = self.group.random_uniform(n_samples=n_samples)
            n_point_2 = self.group.random_uniform(n_samples=n_samples)
            n_point_1 = self.group.regularize(n_point_1)
            n_point_2 = self.group.regularize(n_point_2)

            # Identity and n points 2
            result = metric.dist(point_id, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            if geomstats.tests.tf_backend():
                break
            # n points 1 and identity
            result = metric.dist(n_point_1, point_id)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # one point 1 and n points 2
            result = metric.dist(one_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and one point 2
            result = metric.dist(n_point_1, one_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

            # n points 1 and n points 2
            result = metric.dist(n_point_1, n_point_2)
            self.assertAllClose(gs.shape(result), (n_samples,))

    def test_geodesic_and_belongs(self):
        initial_point = self.group.random_uniform()
        initial_tangent_vec = gs.array([2., 0., -1., 0., 2., 3.])
        metric = self.metrics_all['left_canonical']
        geodesic = metric.geodesic(initial_point=initial_point,
                                   initial_tangent_vec=initial_tangent_vec)

        t = gs.linspace(start=0., stop=1., num=100)
        points = geodesic(t)
        result = gs.all(self.group.belongs(points))
        expected = True
        self.assertAllClose(result, expected)

    def test_geodesic_subsample(self):
        initial_point = gs.array([-1.1, 0., 0.99, 10., 2., 3.])
        initial_tangent_vec = gs.array([1., 0., 2., 1., 1., 1.])
        metric = self.metrics_all['left_canonical']
        geodesic = metric.geodesic(initial_point=initial_point,
                                   initial_tangent_vec=initial_tangent_vec)
        n_steps = 10
        t = gs.linspace(start=0., stop=1., num=n_steps + 1)
        points = geodesic(t)

        tangent_vec_step = initial_tangent_vec / n_steps
        for i in range(n_steps + 1):
            point_step = metric.exp(
                tangent_vec=i * tangent_vec_step,
                base_point=initial_point)
            result = point_step
            expected = points[i]
            self.assertAllClose(result, expected)



"""Checks and associated errors."""

import math

import geomstats.backend as gs


def check_integer(n, n_name):
    """Raise an error if n is not a > 0 integer.

    Parameters
    ----------
    n : unspecified
       Parameter to be tested.
    n_name : string
       Name of the parameter.
    """
    if not(isinstance(n, int) and n > 0):
        if n is not None and n != math.inf:
            raise ValueError(
                '{} is required to be either'
                ' None, math.inf or a strictly positive integer,'
                ' got {}.'.format(n_name, n))


def check_belongs(point, manifold, **kwargs):
    """Raise an error if point does not belong to the input manifold.

    Parameters
    ----------
    point : array-like
        Point to be tested.
    manifold : Manifold
        Manifold to which the point should belong.
    manifold_name : string
        Name of the manifold for the error message.
    """
    if not gs.all(manifold.belongs(point, **kwargs)):
        raise RuntimeError(
            'Some points do not belong to manifold \'%s\' of dimension %d.'
            % (type(manifold).__name__, manifold.dim))


def check_parameter_accepted_values(param, param_name, accepted_values):
    """Raise an error if parameter does not belong to a set of values.

    Parameters
    ----------
    param : unspecified
        Parameter to be tested.
    param_name : string
        Name of the parameter.
    accepted_values : list
        Accepted values that the parameter can take.
    """
    if param not in accepted_values:
        raise ValueError(
            'Parameter {} needs to be in {}, got: {}'.format(
                param_name, accepted_values, param))



"""Prepare and process graph-structured data."""

import logging
import random

import geomstats.backend as gs
from geomstats.geometry.poincare_ball import PoincareBall


class Graph:
    """Class for generating a graph object from a dataset.

    Prepare Graph object from a dataset file.

    Parameters
    ----------
    graph_matrix_path : string
        Path to graph adjacency matrix.
    labels_path : string
        Path to labels of the nodes of the graph.

    Attributes
    ----------
    edges : dict
        Dictionary with node number as key
        and edge connected node numbers as values.
    n_nodes : int
        Number of nodes in the graph.
    labels : dict
        Dictionary with node number as key and the true label number as values.
    """

    def __init__(self, graph_matrix_path, labels_path):
        self.edges = {}
        with open(graph_matrix_path, 'r') as edges_file:
            for i, line in enumerate(edges_file):
                lsp = line.split()
                self.edges[i] = [k for k, value in
                                 enumerate(lsp) if int(value) == 1]

        self.n_nodes = len(self.edges)

        if labels_path is not None:
            self.labels = {}
            with open(labels_path, 'r') as labels_file:
                for i, line in enumerate(labels_file):
                    self.labels[i] = []
                    self.labels[i].append(int(line))

    def random_walk(self, walk_length=5, n_walks_per_node=1):
        """Compute a set of random walks on a graph.

        For each node of the graph, generates a a number of
        random walks of a specified length.
        Two consecutive nodes in the random walk, are necessarily
        related with an edge. The walks capture the structure of the graph.

        Parameters
        ----------
        walk_length : int
            Length of a random walk in terms of number of edges.
        n_walks_per_node : int
            Number of generated walks starting from each node of the graph.

        Returns
        -------
        self : array-like,
            Shape=[n_walks_per_node*self.n_edges), walk_length]
            array containing random walks.
        """
        paths = [[0] * (walk_length + 1)
                 for i in range(self.n_nodes * n_walks_per_node)]

        for index in range(len(self.edges)):
            for i in range(n_walks_per_node):
                paths[index * n_walks_per_node + i] =\
                    self._walk(index, walk_length)
        return gs.array(paths)

    def _walk(self, index, walk_length):
        """Generate a single random walk."""
        count_index = index
        path = [index]
        for _ in range(walk_length):
            count_index = self.edges[count_index][random.randint(
                0, len(self.edges[count_index]) - 1)]
            path.append(count_index)
        return gs.array(path, dtype=gs.int32)


class HyperbolicEmbedding:
    """Class for learning embeddings of graphs on hyperbolic space.

    Parameters
    ----------
    dim : object
        Dimensions of the used hyperbolic space.
    max_epochs : int
        Maximum number of iterations for embedding.
    lr : int
        Learning rate for embedding.
    n_context : int
        Number of nodes to consider from a neighborhood
        of nodes around a particular node.
    n_negative : int
        Number of nodes to consider when searching for
        a set of nodes that are far from a particular node.
    """

    def __init__(
            self, dim=2, max_epochs=100,
            lr=.05, n_context=1, n_negative=2):

        self.manifold = PoincareBall(dim)
        self.max_epochs = max_epochs
        self.lr = lr
        self.n_context = n_context
        self.n_negative = n_negative

    @staticmethod
    def log_sigmoid(vector):
        """Logsigmoid function.

        Apply log sigmoid function.

        Parameters
        ----------
        vector : array-like, shape=[n_samples, dim]

        Returns
        -------
        result : array-like, shape=[n_samples, dim]
        """
        return gs.log((1 / (1 + gs.exp(-vector))))

    @staticmethod
    def grad_log_sigmoid(vector):
        """Gradient of log sigmoid function.

        Parameters
        ----------
        vector : array-like, shape=[n_samples, dim]

        Returns
        -------
        gradient : array-like, shape=[n_samples, dim]
        """
        return 1 / (1 + gs.exp(vector))

    def grad_squared_distance(self, point_a, point_b):
        """Gradient of squared hyperbolic distance.

        Gradient of the squared distance based on the
        Ball representation according to point_a.

        Parameters
        ----------
        point_a : array-like, shape=[n_samples, dim]
            First point in hyperbolic space.
        point_b : array-like, shape=[n_samples, dim]
            Second point in hyperbolic space.

        Returns
        -------
        dist : array-like, shape=[n_samples, 1]
            Geodesic squared distance between the two points.
        """
        hyperbolic_metric = self.manifold.metric
        log_map = hyperbolic_metric.log(point_b, point_a)

        return -2 * log_map

    def loss(
            self, example_embedding, context_embedding, negative_embedding):
        """Compute loss and grad.

        Compute loss and grad given embedding of the current example,
        embedding of the context and negative sampling embedding.

        Parameters
        ----------
        example_embedding : array-like, shape=[dim]
            Current data sample embedding.
        context_embedding : array-like, shape=[dim]
            Current context embedding.
        negative_embedding: array-like, shape=[dim]
            Current negative sample embedding.

        Returns
        -------
        total_loss : int
            The current value of the loss function.
        example_grad : array-like, shape=[dim]
            The gradient of the loss function at the embedding
            of the current data sample.
        """
        n_edges, dim =\
            negative_embedding.shape[0], example_embedding.shape[-1]
        example_embedding = gs.expand_dims(example_embedding, 0)
        context_embedding = gs.expand_dims(context_embedding, 0)

        positive_distance =\
            self.manifold.metric.squared_dist(
                example_embedding, context_embedding)
        positive_loss =\
            self.log_sigmoid(-positive_distance)

        reshaped_example_embedding =\
            gs.repeat(example_embedding, n_edges, axis=0)

        negative_distance =\
            self.manifold.metric.squared_dist(
                reshaped_example_embedding, negative_embedding)
        negative_loss = self.log_sigmoid(negative_distance)

        total_loss = -(positive_loss + gs.sum(negative_loss))

        positive_log_sigmoid_grad =\
            -self.grad_log_sigmoid(-positive_distance)

        positive_distance_grad =\
            self.grad_squared_distance(example_embedding, context_embedding)

        positive_grad =\
            gs.repeat(positive_log_sigmoid_grad, dim, axis=-1)\
            * positive_distance_grad

        negative_distance_grad =\
            self.grad_squared_distance(
                reshaped_example_embedding, negative_embedding)

        negative_distance = gs.to_ndarray(negative_distance,
                                          to_ndim=2, axis=-1)
        negative_log_sigmoid_grad =\
            self.grad_log_sigmoid(negative_distance)

        negative_grad = negative_log_sigmoid_grad\
            * negative_distance_grad

        example_grad = -(positive_grad + gs.sum(negative_grad, axis=0))

        return total_loss, example_grad

    def embed(self, graph):
        """Compute embedding.

        Optimize a loss function to obtain a representable embedding.

        Parameters
        ----------
        graph : object
            An instance of the Graph class.

        Returns
        -------
        embeddings : array-like, shape=[n_samples, dim]
            Return the embedding of the data. Each data sample
            is represented as a point belonging to the manifold.
        """
        nb_vertices_by_edges = \
            [len(e_2) for _, e_2 in graph.edges.items()]
        logging.info('Number of edges: %s', len(graph.edges))
        logging.info(
            'Mean vertices by edges: %s',
            (sum(nb_vertices_by_edges, 0) / len(graph.edges)))

        negative_table_parameter = 5
        negative_sampling_table = []

        for i, nb_v in enumerate(nb_vertices_by_edges):
            negative_sampling_table += \
                ([i] * int((nb_v ** (3. / 4.))) * negative_table_parameter)

        negative_sampling_table = gs.array(negative_sampling_table)
        random_walks = graph.random_walk()
        embeddings = gs.random.normal(
            size=(graph.n_nodes, self.manifold.dim))
        embeddings = embeddings * 0.2

        for epoch in range(self.max_epochs):
            total_loss = []
            for path in random_walks:

                for example_index, one_path in enumerate(path):
                    context_index = path[max(
                        0, example_index - self.n_context):
                            min(example_index + self.n_context,
                                len(path))]
                    negative_index = \
                        gs.random.randint(negative_sampling_table.shape[0],
                                          size=(len(context_index),
                                                self.n_negative))

                    negative_index = gs.expand_dims(
                        gs.flatten(negative_index), axis=-1)

                    negative_index = gs.get_slice(
                        negative_sampling_table, negative_index)

                    example_embedding = embeddings[gs.cast(
                        one_path, dtype=gs.int64)]

                    for one_context_i, one_negative_i in zip(context_index,
                                                             negative_index):
                        context_embedding = embeddings[one_context_i]

                        negative_embedding = gs.get_slice(
                            embeddings, gs.squeeze(gs.cast(
                                one_negative_i, dtype=gs.int64)))

                        l, g_ex = self.loss(
                            example_embedding,
                            context_embedding,
                            negative_embedding)
                        total_loss.append(l)

                        example_to_update = embeddings[one_path]

                        valeur = self.manifold.metric.exp(
                            -self.lr * g_ex, example_to_update)

                        embeddings = gs.assignment(
                            embeddings, valeur, gs.to_ndarray(
                                one_path, to_ndim=1), axis=1)

            logging.info(
                'iteration %d loss_value %f',
                epoch, sum(total_loss, 0) / len(total_loss))
        return embeddings



"""Unit tests for the KDE classifier."""

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.hypersphere import Hypersphere
from geomstats.geometry.poincare_ball import PoincareBall
from geomstats.learning.kernel_density_estimation_classifier \
    import KernelDensityEstimationClassifier
from geomstats.learning.radial_kernel_functions import triangular_radial_kernel

TOLERANCE = 1e-4


class TestKernelDensityEstimationClassifier(geomstats.tests.TestCase):
    """Class defining the Kernel Density Estimation Classifier tests."""

    def setUp(self):
        """Define the parameters to test."""
        gs.random.seed(1234)
        self.dim = 2
        self.space = Euclidean(dim=self.dim)
        self.distance = self.space.metric.dist

    def test_predict(self):
        """Test the 'predict' class method."""
        training_dataset = gs.array(
            [[0.0, 0.0],
             [1.0, 0.0],
             [2.0, 0.0],
             [3.0, 0.0]])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            distance=self.distance)
        kde.fit(training_dataset, labels)
        result = kde.predict(gs.array([[1.1, 0.0]]))
        expected = gs.array([0])
        self.assertAllClose(expected, result)

    def test_predict_one_dimensional_data(self):
        """Test the 'predict' class method."""
        training_dataset = gs.array(
            [[0.0],
             [1.0],
             [2.0],
             [3.0]])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            distance='minkowski')
        kde.fit(training_dataset, labels)
        result = kde.predict(gs.array([1.1]))
        expected = gs.array([0])
        self.assertAllClose(expected, result)

    @geomstats.tests.np_only
    def test_predict_one_dimensional_data_callable_distance(self):
        """Test the 'predict' class method on one dimensional data."""
        training_dataset = gs.array([0, 1, 2, 3])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            distance=self.distance)
        kde.fit(training_dataset, labels)
        result = kde.predict(gs.array([1.1]))
        expected = gs.array([0])
        self.assertAllClose(expected, result)

    @geomstats.tests.np_only
    def test_predict_proba_uniform_kernel_one_dimensional_data(self):
        """Test the 'predict_proba' class method using the 'uniform' kernel.

        Test the 'predict_proba' class method using the 'uniform' kernel on
        one-dimensional date of shape [n_samples,].
        """
        training_dataset = gs.array([0, 1, 2, 3])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            kernel='uniform',
            distance=self.distance)
        kde.fit(training_dataset, labels)
        result = kde.predict_proba(gs.array([0.9]))
        expected = gs.array([[1 / 2, 1 / 2]])
        self.assertAllClose(expected, result, atol=TOLERANCE)

    def test_predict_proba_uniform_kernel(self):
        """Test the 'predict_proba' class method using the 'uniform' kernel."""
        training_dataset = gs.array(
            [[0.0, 0.0],
             [1.0, 0.0],
             [2.0, 0.0],
             [3.0, 0.0]])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            kernel='uniform',
            distance=self.distance)
        kde.fit(training_dataset, labels)
        result = kde.predict_proba(gs.array([[0.9, 0.0]]))
        expected = gs.array([[1 / 2, 1 / 2]])
        self.assertAllClose(expected, result, atol=TOLERANCE)

    def test_predict_proba_distance_kernel(self):
        """Test the 'predict_proba' class method using 'distance' kernel."""
        training_dataset = gs.array(
            [[0.0, 0.0],
             [1.0, 0.0],
             [2.0, 0.0],
             [3.0, 0.0]])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            kernel='distance',
            distance=self.distance)
        kde.fit(training_dataset, labels)
        result = kde.predict_proba(gs.array([[1.0, 0.0]]))
        expected = gs.array([[1, 0]])
        self.assertAllClose(expected, result, atol=TOLERANCE)

    @geomstats.tests.np_and_pytorch_only
    def test_predict_proba_triangular_kernel(self):
        """Test the 'predict_proba' class method using a triangular kernel."""
        training_dataset = gs.array(
            [[0.0, 0.0],
             [1.0, 0.0],
             [2.0, 0.0],
             [3.0, 0.0]])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            kernel=triangular_radial_kernel,
            bandwidth=2.0,
            p=2,
            distance='minkowski')
        kde.fit(training_dataset, labels)
        result = kde.predict_proba(gs.array([[1.0, 0.0]]))
        expected = gs.array([[3 / 4, 1 / 4]])
        self.assertAllClose(expected, result, atol=TOLERANCE)

    @geomstats.tests.np_and_pytorch_only
    def test_predict_proba_triangular_kernel_callable_distance(self):
        """Test the 'predict_proba' class method using a triangular kernel."""
        training_dataset = gs.array(
            [[0.0, 0.0],
             [1.0, 0.0],
             [2.0, 0.0],
             [3.0, 0.0]])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            kernel=triangular_radial_kernel,
            bandwidth=2.0,
            distance=self.distance)
        kde.fit(training_dataset, labels)
        result = kde.predict_proba(gs.array([[1.0, 0.0]]))
        expected = gs.array([[3 / 4, 1 / 4]])
        self.assertAllClose(expected, result, atol=TOLERANCE)

    @geomstats.tests.np_and_pytorch_only
    def test_predict_triangular_kernel_callable_distance(self):
        """Test the 'predict' class method using a triangular kernel."""
        training_dataset = gs.array(
            [[0.0, 0.0],
             [1.0, 0.0],
             [2.0, 0.0],
             [3.0, 0.0]])
        labels = [0, 0, 1, 1]
        kde = KernelDensityEstimationClassifier(
            kernel=triangular_radial_kernel,
            bandwidth=2.0,
            distance=self.distance)
        kde.fit(training_dataset, labels)
        result = kde.predict(gs.array([[1.0, 0.0], [1.0, 0.0]]))
        expected = gs.array([0, 0])
        self.assertAllClose(expected, result, atol=TOLERANCE)

    def test_predict_hypersphere_distance(self):
        """Test the 'predict' class method using the hypersphere distance."""
        dim = 2
        space = Hypersphere(dim=dim)
        distance = space.metric.dist
        training_dataset = gs.array(
            [[1, 0, 0],
             [3 ** (1 / 2) / 2, 1 / 2, 0],
             [3 ** (1 / 2) / 2, - 1 / 2, 0],
             [0, 0, 1],
             [0, 1 / 2, 3 ** (1 / 2) / 2],
             [0, - 1 / 2, 3 ** (1 / 2) / 2]])
        labels = [0, 0, 0, 1, 1, 1]
        kde = KernelDensityEstimationClassifier(
            distance=distance)
        kde.fit(training_dataset, labels)
        target_dataset = gs.array(
            [[2 ** (1 / 2) / 2, 2 ** (1 / 2) / 2, 0],
             [0, 1 / 2, - 3 ** (1 / 2) / 2],
             [0, - 1 / 2, - 3 ** (1 / 2) / 2],
             [- 3 ** (1 / 2) / 2, 1 / 2, 0],
             [- 3 ** (1 / 2) / 2, - 1 / 2, 0],
             [0, 2 ** (1 / 2) / 2, 2 ** (1 / 2) / 2]])
        result = kde.predict(target_dataset)
        expected = [0, 0, 0, 1, 1, 1]
        self.assertAllClose(expected, result)

    def test_predict_poincare_ball_distance(self):
        """Test the 'predict' class method using the Poincare ball distance."""
        dim = 2
        space = PoincareBall(dim=dim)
        distance = space.metric.dist
        training_dataset = gs.array(
            [[1 / 2, 1 / 4],
             [1 / 2, 0],
             [1 / 2, - 1 / 4],
             [- 1 / 2, 1 / 4],
             [- 1 / 2, 0],
             [- 1 / 2, - 1 / 4]])
        labels = [0, 0, 0, 1, 1, 1]
        kde = KernelDensityEstimationClassifier(
            distance=distance,
            kernel='distance')
        kde.fit(training_dataset, labels)
        target_dataset = gs.array(
            [[1 / 2, 1 / 5],
             [1 / 2, 0],
             [1 / 2, - 1 / 5],
             [- 1 / 2, 1 / 5],
             [- 1 / 2, 0],
             [- 1 / 2, - 1 / 5]])
        result = kde.predict(target_dataset)
        expected = [0, 0, 0, 1, 1, 1]
        self.assertAllClose(expected, result)

    def test_predict_hyperboloid_distance(self):
        """Test the 'predict' class method using the hyperboloid distance."""
        dim = 2
        space = Hyperboloid(dim=dim)
        distance = space.metric.dist
        training_dataset_intrinsic = gs.array(
            [[1 / 2, 1 / 4],
             [1 / 2, 0],
             [1 / 2, - 1 / 4],
             [- 1 / 2, 1 / 4],
             [- 1 / 2, 0],
             [- 1 / 2, - 1 / 4]])
        training_dataset = space.change_coordinates_system(
            training_dataset_intrinsic,
            from_coordinates_system='intrinsic',
            to_coordinates_system='extrinsic')
        labels = [0, 0, 0, 1, 1, 1]
        kde = KernelDensityEstimationClassifier(
            distance=distance,
            kernel='distance')
        kde.fit(training_dataset, labels)
        target_dataset_intrinsic = gs.array(
            [[1 / 2, 1 / 5],
             [1 / 2, 0],
             [1 / 2, - 1 / 5],
             [- 1 / 2, 1 / 5],
             [- 1 / 2, 0],
             [- 1 / 2, - 1 / 5]])
        target_dataset = space.change_coordinates_system(
            target_dataset_intrinsic,
            from_coordinates_system='intrinsic',
            to_coordinates_system='extrinsic')
        result = kde.predict(target_dataset)
        expected = [0, 0, 0, 1, 1, 1]
        self.assertAllClose(expected, result)



"""Unit tests for Tangent PCA."""

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.spd_matrices import SPDMatrices, SPDMetricAffine
from geomstats.geometry.special_euclidean import SpecialEuclidean
from geomstats.geometry.special_orthogonal import SpecialOrthogonal
from geomstats.learning.exponential_barycenter import ExponentialBarycenter
from geomstats.learning.pca import TangentPCA


class TestTangentPCA(geomstats.tests.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        self.so3 = SpecialOrthogonal(n=3, point_type='vector')
        self.spd = SPDMatrices(3)
        self.spd_metric = SPDMetricAffine(3)

        self.n_samples = 10

        self.X = self.so3.random_uniform(n_samples=self.n_samples)
        self.metric = self.so3.bi_invariant_metric
        self.n_components = 2

    @geomstats.tests.np_only
    def test_tangent_pca_error(self):
        X = self.X
        tpca = TangentPCA(self.metric, n_components=self.n_components)
        tpca.fit(X)
        X_diff_size = gs.ones((self.n_samples, gs.shape(X)[1] + 1))
        self.assertRaises(ValueError, tpca.transform, X_diff_size)

    @geomstats.tests.np_only
    def test_tangent_pca(self):
        X = self.X
        tpca = TangentPCA(self.metric, n_components=gs.shape(X)[1])
        tpca.fit(X)
        self.assertEqual(tpca.n_features_, gs.shape(X)[1])

    @geomstats.tests.np_only
    def test_fit_mle(self):
        X = self.X
        tpca = TangentPCA(self.metric, n_components='mle')
        tpca.fit(X)
        self.assertEqual(tpca.n_features_, gs.shape(X)[1])

    @geomstats.tests.np_only
    def test_fit_to_target_explained_variance(self):
        X = self.spd.random_uniform(n_samples=5)
        target = 0.90
        tpca = TangentPCA(
            self.spd_metric, n_components=target)
        tpca.fit(X)
        result = gs.cumsum(tpca.explained_variance_ratio_)[-1] > target
        expected = True
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_fit_matrix(self):
        expected = 2
        X = self.spd.random_uniform(n_samples=5)
        tpca = TangentPCA(
            metric=self.spd_metric, n_components=expected)
        tpca.fit(X)
        result = tpca.n_components_
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_fit_transform_matrix(self):
        expected = 2
        X = self.spd.random_uniform(n_samples=5)
        tpca = TangentPCA(
            metric=self.spd_metric, n_components=expected)
        tangent_projected_data = tpca.fit_transform(X)
        result = tangent_projected_data.shape[-1]
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_fit_inverse_transform_matrix(self):
        X = self.spd.random_uniform(n_samples=5)
        tpca = TangentPCA(
            metric=self.spd_metric)
        tangent_projected_data = tpca.fit_transform(X)
        result = tpca.inverse_transform(tangent_projected_data)
        expected = X
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_fit_transform_vector(self):
        expected = 2
        tpca = TangentPCA(
            metric=self.metric, n_components=expected)
        tangent_projected_data = tpca.fit_transform(self.X)
        result = tangent_projected_data.shape[-1]
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_fit_inverse_transform_vector(self):
        tpca = TangentPCA(metric=self.metric)
        tangent_projected_data = tpca.fit_transform(self.X)
        result = tpca.inverse_transform(tangent_projected_data)
        expected = self.X
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_fit_fit_transform_matrix(self):
        X = self.spd.random_uniform(n_samples=5)
        tpca = TangentPCA(
            metric=self.spd_metric)
        expected = tpca.fit_transform(X)
        result = tpca.fit(X).transform(X)
        self.assertAllClose(result, expected)

    @geomstats.tests.np_only
    def test_fit_matrix_se(self):
        se_mat = SpecialEuclidean(n=3)
        X = se_mat.random_uniform(self.n_samples)
        estimator = ExponentialBarycenter(se_mat)
        estimator.fit(X)
        mean = estimator.estimate_
        tpca = TangentPCA(metric=se_mat)
        tangent_projected_data = tpca.fit_transform(X, base_point=mean)
        result = tpca.inverse_transform(tangent_projected_data)
        expected = X
        self.assertAllClose(result, expected)



"""Unit tests for the Euclidean space."""

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.euclidean import Euclidean


class TestEuclidean(geomstats.tests.TestCase):
    def setUp(self):
        gs.random.seed(1234)

        self.dimension = 2
        self.space = Euclidean(self.dimension)
        self.metric = self.space.metric

        self.n_samples = 3

        self.one_point_a = gs.array([0., 1.])
        self.one_point_b = gs.array([2., 10.])
        self.n_points_a = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])
        self.n_points_b = gs.array([
            [2., 10.],
            [8., -1.],
            [-3., 6.]])

    def test_random_uniform_and_belongs(self):
        point = self.space.random_uniform()
        result = self.space.belongs(point)
        expected = True

        self.assertAllClose(result, expected)

    def test_squared_norm_vectorization(self):
        n_samples = self.n_samples
        n_points = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])
        result = self.metric.squared_norm(n_points)

        expected = gs.array([5., 20., 26.])

        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

    def test_norm_vectorization_single_sample(self):
        one_point = gs.array([[0., 1.]])

        result = self.metric.norm(one_point)
        expected = gs.array([1.])
        self.assertAllClose(gs.shape(result), (1,))
        self.assertAllClose(result, expected)

        one_point = gs.array([0., 1.])

        result = self.metric.norm(one_point)
        expected = 1.
        self.assertAllClose(gs.shape(result), ())
        self.assertAllClose(result, expected)

    def test_norm_vectorization_n_samples(self):
        n_samples = self.n_samples
        n_points = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])

        result = self.metric.norm(n_points)

        expected = gs.array([2.2360679775, 4.472135955, 5.09901951359])

        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

    def test_exp_vectorization(self):
        n_samples = self.n_samples
        dim = self.dimension

        one_tangent_vec = gs.array([0., 1.])
        one_base_point = gs.array([2., 10.])
        n_tangent_vecs = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])
        n_base_points = gs.array([
            [2., 10.],
            [8., -1.],
            [-3., 6.]])

        result = self.metric.exp(one_tangent_vec, one_base_point)
        expected = one_tangent_vec + one_base_point

        self.assertAllClose(result, expected)

        result = self.metric.exp(n_tangent_vecs, one_base_point)
        self.assertAllClose(gs.shape(result), (n_samples, dim))

        result = self.metric.exp(one_tangent_vec, n_base_points)
        self.assertAllClose(gs.shape(result), (n_samples, dim))

        result = self.metric.exp(n_tangent_vecs, n_base_points)
        self.assertAllClose(gs.shape(result), (n_samples, dim))

    def test_log_vectorization(self):
        n_samples = self.n_samples
        dim = self.dimension

        one_point = gs.array([0., 1.])
        one_base_point = gs.array([2., 10.])
        n_points = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])
        n_base_points = gs.array([
            [2., 10.],
            [8., -1.],
            [-3., 6.]])

        result = self.metric.log(one_point, one_base_point)
        expected = one_point - one_base_point
        self.assertAllClose(result, expected)

        result = self.metric.log(n_points, one_base_point)
        self.assertAllClose(gs.shape(result), (n_samples, dim))

        result = self.metric.log(one_point, n_base_points)
        self.assertAllClose(gs.shape(result), (n_samples, dim))

        result = self.metric.log(n_points, n_base_points)
        self.assertAllClose(gs.shape(result), (n_samples, dim))

    def test_squared_dist_vectorization(self):
        n_samples = self.n_samples

        one_point_a = gs.array([0., 1.])
        one_point_b = gs.array([2., 10.])
        n_points_a = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])
        n_points_b = gs.array([
            [2., 10.],
            [8., -1.],
            [-3., 6.]])

        result = self.metric.squared_dist(one_point_a, one_point_b)
        vec = one_point_a - one_point_b
        expected = gs.dot(vec, gs.transpose(vec))
        self.assertAllClose(result, expected)

        result = self.metric.squared_dist(n_points_a, one_point_b)
        self.assertAllClose(gs.shape(result), (n_samples,))

        result = self.metric.squared_dist(one_point_a, n_points_b)
        self.assertAllClose(gs.shape(result), (n_samples,))

        result = self.metric.squared_dist(n_points_a, n_points_b)
        expected = gs.array([81., 109., 29.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

    def test_dist_vectorization(self):
        n_samples = self.n_samples

        one_point_a = gs.array([0., 1.])
        one_point_b = gs.array([2., 10.])
        n_points_a = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])
        n_points_b = gs.array([
            [2., 10.],
            [8., -1.],
            [-3., 6.]])

        result = self.metric.dist(one_point_a, one_point_b)
        vec = one_point_a - one_point_b
        expected = gs.sqrt(gs.dot(vec, gs.transpose(vec)))
        self.assertAllClose(result, expected)

        result = self.metric.dist(n_points_a, one_point_b)
        self.assertAllClose(gs.shape(result), (n_samples,))

        result = self.metric.dist(one_point_a, n_points_b)
        self.assertAllClose(gs.shape(result), (n_samples,))

        result = self.metric.dist(n_points_a, n_points_b)
        expected = gs.array([9., gs.sqrt(109.), gs.sqrt(29.)])

        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

    def test_belongs(self):
        point = gs.array([0., 1.])

        result = self.space.belongs(point)
        expected = True

        self.assertAllClose(result, expected)

    def test_random_uniform(self):
        result = self.space.random_uniform()

        self.assertAllClose(gs.shape(result), (self.dimension,))

    def test_inner_product_matrix(self):
        result = self.metric.inner_product_matrix()

        expected = gs.eye(self.dimension)

        self.assertAllClose(result, expected)

    def test_inner_product(self):
        point_a = gs.array([0., 1.])
        point_b = gs.array([2., 10.])

        result = self.metric.inner_product(point_a, point_b)
        expected = 10.

        self.assertAllClose(result, expected)

    def test_inner_product_vectorization_single_sample(self):
        one_point_a = gs.array([[0., 1.]])
        one_point_b = gs.array([[2., 10.]])

        result = self.metric.inner_product(one_point_a, one_point_b)
        expected = gs.array([10.])
        self.assertAllClose(gs.shape(result), (1,))
        self.assertAllClose(result, expected)

        one_point_a = gs.array([[0., 1.]])
        one_point_b = gs.array([2., 10.])

        result = self.metric.inner_product(one_point_a, one_point_b)
        expected = gs.array([10.])
        self.assertAllClose(gs.shape(result), (1,))
        self.assertAllClose(result, expected)

        one_point_a = gs.array([0., 1.])
        one_point_b = gs.array([[2., 10.]])

        result = self.metric.inner_product(one_point_a, one_point_b)
        expected = gs.array([10.])
        self.assertAllClose(gs.shape(result), (1,))
        self.assertAllClose(result, expected)

        one_point_a = gs.array([0., 1.])
        one_point_b = gs.array([2., 10.])

        result = self.metric.inner_product(one_point_a, one_point_b)
        expected = 10.
        self.assertAllClose(gs.shape(result), ())
        self.assertAllClose(result, expected)

    def test_inner_product_vectorization_n_samples(self):
        n_samples = 3
        n_points_a = gs.array([
            [2., 1.],
            [-2., -4.],
            [-5., 1.]])
        n_points_b = gs.array([
            [2., 10.],
            [8., -1.],
            [-3., 6.]])

        one_point_a = gs.array([0., 1.])
        one_point_b = gs.array([2., 10.])

        result = self.metric.inner_product(n_points_a, one_point_b)
        expected = gs.array([14., -44., 0.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(one_point_a, n_points_b)
        expected = gs.array([10., -1., 6.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(n_points_a, n_points_b)
        expected = gs.array([14., -12., 21.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        one_point_a = gs.array([[0., 1.]])
        one_point_b = gs.array([[2., 10]])

        result = self.metric.inner_product(n_points_a, one_point_b)
        expected = gs.array([14., -44., 0.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(one_point_a, n_points_b)
        expected = gs.array([10., -1., 6.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(n_points_a, n_points_b)
        expected = gs.array([14., -12., 21.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        one_point_a = gs.array([[0., 1.]])
        one_point_b = gs.array([2., 10.])

        result = self.metric.inner_product(n_points_a, one_point_b)
        expected = gs.array([14., -44., 0.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(one_point_a, n_points_b)
        expected = gs.array([10., -1., 6.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(n_points_a, n_points_b)
        expected = gs.array([14., -12., 21.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        one_point_a = gs.array([0., 1.])
        one_point_b = gs.array([[2., 10.]])

        result = self.metric.inner_product(n_points_a, one_point_b)
        expected = gs.array([14., -44., 0.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(one_point_a, n_points_b)
        expected = gs.array([10., -1., 6.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

        result = self.metric.inner_product(n_points_a, n_points_b)
        expected = gs.array([14., -12., 21.])
        self.assertAllClose(gs.shape(result), (n_samples,))
        self.assertAllClose(result, expected)

    def test_squared_norm(self):
        point = gs.array([-2., 4.])

        result = self.metric.squared_norm(point)
        expected = 20.

        self.assertAllClose(result, expected)

    def test_norm(self):
        point = gs.array([-2., 4.])
        result = self.metric.norm(point)
        expected = 4.472135955

        self.assertAllClose(result, expected)

    def test_exp(self):
        base_point = gs.array([0., 1.])
        vector = gs.array([2., 10.])

        result = self.metric.exp(tangent_vec=vector,
                                 base_point=base_point)
        expected = base_point + vector

        self.assertAllClose(result, expected)

    def test_log(self):
        base_point = gs.array([0., 1.])
        point = gs.array([2., 10.])

        result = self.metric.log(point=point, base_point=base_point)
        expected = point - base_point

        self.assertAllClose(result, expected)

    def test_squared_dist(self):
        point_a = gs.array([-1., 4.])
        point_b = gs.array([1., 1.])

        result = self.metric.squared_dist(point_a, point_b)
        vec = point_b - point_a
        expected = gs.dot(vec, vec)

        self.assertAllClose(result, expected)

    def test_dist(self):
        point_a = gs.array([0., 1.])
        point_b = gs.array([2., 10.])

        result = self.metric.dist(point_a, point_b)
        expected = gs.linalg.norm(point_b - point_a)

        self.assertAllClose(result, expected)

    def test_geodesic_and_belongs(self):
        n_geodesic_points = 100
        initial_point = gs.array([[2., -1.]])
        initial_tangent_vec = gs.array([2., 0.])
        geodesic = self.metric.geodesic(
            initial_point=initial_point,
            initial_tangent_vec=initial_tangent_vec)

        t = gs.linspace(start=0., stop=1., num=n_geodesic_points)
        points = geodesic(t)

        result = self.space.belongs(points)
        expected = gs.array(n_geodesic_points * [True])

        self.assertAllClose(expected, result)



"""Unit tests for the Hyperbolic space coordinates change.

We verify poincare ball model, poincare half plane
and minkowisky extrinsic/intrisic.
We also verify that converting point will lead to get same
distance (implemented for ball model and extrinsic only)
"""
import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.hyperbolic import Hyperbolic
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.poincare_ball import PoincareBall


class TestHyperbolicCoords(geomstats.tests.TestCase):
    def setUp(self):
        gs.random.seed(1234)
        self.dimension = 2

        self.extrinsic_manifold = Hyperboloid(
            dim=self.dimension)
        self.extrinsic_metric = self.extrinsic_manifold.metric

        self.ball_manifold = PoincareBall(
            dim=self.dimension)
        self.ball_metric = self.ball_manifold.metric

        self.intrinsic_manifold = Hyperboloid(
            dim=self.dimension, coords_type='intrinsic')
        self.intrinsic_metric = self.intrinsic_manifold.metric

        self.n_samples = 10

    def test_extrinsic_ball_extrinsic(self):
        x_in = gs.array([0.5, 7])
        x = self.intrinsic_manifold.to_coordinates(
            x_in, to_coords_type='extrinsic')
        x_b = self.extrinsic_manifold.to_coordinates(x, to_coords_type='ball')
        x2 = self.ball_manifold.to_coordinates(x_b, to_coords_type='extrinsic')
        self.assertAllClose(x, x2, atol=1e-8)

    def test_belongs_intrinsic(self):
        x_in = gs.array([0.5, 7])
        is_in = self.intrinsic_manifold.belongs(x_in)
        self.assertTrue(is_in)

    def test_belongs_extrinsic(self):
        x_true = self.intrinsic_manifold.to_coordinates(
            gs.array([0.5, 7]), 'extrinsic')
        x_false = gs.array([0.5, 7, 3.])
        is_in = self.extrinsic_manifold.belongs(x_true)
        self.assertTrue(is_in)
        is_out = self.extrinsic_manifold.belongs(x_false)
        self.assertFalse(is_out)

    def test_belongs_ball(self):
        x_true = gs.array([0.5, 0.5])
        x_false = gs.array([0.8, 0.8])
        is_in = self.ball_manifold.belongs(x_true)
        self.assertTrue(is_in)
        is_out = self.ball_manifold.belongs(x_false)
        self.assertFalse(is_out)

    def test_extrinsic_half_plane_extrinsic(self):
        x_in = gs.array([0.5, 7])
        x = self.intrinsic_manifold.to_coordinates(
            x_in, to_coords_type='extrinsic')
        x_up = self.extrinsic_manifold.to_coordinates(
            x, to_coords_type='half-plane')

        x2 = Hyperbolic.change_coordinates_system(
            x_up, 'half-plane', 'extrinsic')
        self.assertAllClose(x, x2, atol=1e-8)

    def test_intrinsic_extrinsic_intrinsic(self):
        x_intr = gs.array([0.5, 7])
        x_extr = self.intrinsic_manifold.to_coordinates(
            x_intr, to_coords_type='extrinsic')
        x_intr2 = self.extrinsic_manifold.to_coordinates(
            x_extr, to_coords_type='intrinsic')
        self.assertAllClose(x_intr, x_intr2, atol=1e-8)

    def test_ball_extrinsic_ball(self):
        x = gs.array([0.5, 0.2])
        x_e = self.ball_manifold.to_coordinates(x, to_coords_type='extrinsic')
        x2 = self.extrinsic_manifold.to_coordinates(x_e, to_coords_type='ball')
        self.assertAllClose(x, x2, atol=1e-10)

    def test_distance_ball_extrinsic_from_ball(self):
        x_ball = gs.array([0.7, 0.2])
        y_ball = gs.array([0.2, 0.2])
        x_extr = self.ball_manifold.to_coordinates(
            x_ball, to_coords_type='extrinsic')
        y_extr = self.ball_manifold.to_coordinates(
            y_ball, to_coords_type='extrinsic')
        dst_ball = self.ball_metric.dist(x_ball, y_ball)
        dst_extr = self.extrinsic_metric.dist(x_extr, y_extr)

        self.assertAllClose(dst_ball, dst_extr)

    def test_distance_ball_extrinsic_from_extr(self):
        x_int = gs.array([10, 0.2])
        y_int = gs.array([1, 6.])
        x_extr = self.intrinsic_manifold.to_coordinates(
            x_int, to_coords_type='extrinsic')
        y_extr = self.intrinsic_manifold.to_coordinates(
            y_int, to_coords_type='extrinsic')
        x_ball = self.extrinsic_manifold.to_coordinates(
            x_extr, to_coords_type='ball')
        y_ball = self.extrinsic_manifold.to_coordinates(
            y_extr, to_coords_type='ball')
        dst_ball = self.ball_metric.dist(x_ball, y_ball)
        dst_extr = self.extrinsic_metric.dist(x_extr, y_extr)

        self.assertAllClose(dst_ball, dst_extr)

    def test_distance_ball_extrinsic_from_extr_4_dim(self):
        x_int = gs.array([10, 0.2, 3, 4])
        y_int = gs.array([1, 6, 2., 1])

        ball_manifold = PoincareBall(4)
        extrinsic_manifold = Hyperboloid(4)

        ball_metric = ball_manifold.metric
        extrinsic_metric = extrinsic_manifold.metric

        x_extr = extrinsic_manifold.from_coordinates(
            x_int, from_coords_type='intrinsic')
        y_extr = extrinsic_manifold.from_coordinates(
            y_int, from_coords_type='intrinsic')
        x_ball = extrinsic_manifold.to_coordinates(
            x_extr, to_coords_type='ball')
        y_ball = extrinsic_manifold.to_coordinates(
            y_extr, to_coords_type='ball')
        dst_ball = ball_metric.dist(x_ball, y_ball)
        dst_extr = extrinsic_metric.dist(x_extr, y_extr)

        self.assertAllClose(dst_ball, dst_extr)

    def test_log_exp_ball_extrinsic_from_extr(self):
        """Compare log exp in different parameterizations."""
        x_int = gs.array([4., 0.2])
        y_int = gs.array([3., 3])
        x_extr = self.intrinsic_manifold.to_coordinates(
            x_int, to_coords_type='extrinsic')
        y_extr = self.intrinsic_manifold.to_coordinates(
            y_int, to_coords_type='extrinsic')
        x_ball = self.extrinsic_manifold.to_coordinates(
            x_extr, to_coords_type='ball')
        y_ball = self.extrinsic_manifold.to_coordinates(
            y_extr, to_coords_type='ball')

        x_ball_log_exp = self.ball_metric.exp(
            self.ball_metric.log(y_ball, x_ball), x_ball)

        x_extr_a = self.extrinsic_metric.exp(
            self.extrinsic_metric.log(y_extr, x_extr), x_extr)
        x_extr_b = self.extrinsic_manifold.from_coordinates(
            x_ball_log_exp, from_coords_type='ball')
        self.assertAllClose(x_extr_a, x_extr_b, atol=1e-4)

    def test_log_exp_ball(self):
        x = gs.array([0.1, 0.2])
        y = gs.array([0.2, 0.5])

        log = self.ball_metric.log(point=y, base_point=x)
        exp = self.ball_metric.exp(tangent_vec=log, base_point=x)
        self.assertAllClose(exp, y, atol=1e-1)

    def test_log_exp_ball_vectorization(self):
        x = gs.array([0.1, 0.2])
        y = gs.array([[0.2, 0.5], [0.1, 0.7]])

        log = self.ball_metric.log(y, x)
        exp = self.ball_metric.exp(log, x)
        self.assertAllClose(exp, y, atol=1e-1)

    def test_log_exp_ball_null_tangent(self):
        x = gs.array([[0.1, 0.2], [0.1, 0.2]])
        tangent_vec = gs.array([[0.0, 0.0], [0.0, 0.0]])
        exp = self.ball_metric.exp(tangent_vec, x)
        self.assertAllClose(exp, x, atol=1e-10)
